/******************************************************************************/
/*                                                                            */
/* Header file for the ibd_create subprograms: beaglesim, beagledag,          */
/* simpop_fgl, simped_fgl, fgl2haplo, fgl2ibd and fgl2dgl.                    */
/*                                                                            */
/******************************************************************************/

#ifndef IBD_CREATE_H
#define IBD_CREATE_H

#include "incs.h"

/* Data constructs used by simpop_fgl, fgl2haplo and maybe some of the other  */
/* subroutines associated with these subprograms:                             */

/* Chris Glazner decided to use signed long longs because although an actual  */
/* address is never negative, some addresses get used in arithmetic with      */
/* negative numbers and it was going haywire with unsigned long longs.        */
/* The choice of long vs. long long probably does not matter but 1e8 base     */
/* pairs is not much smaller than the signed long maximum, which is ~2e9.     */

typedef struct _Chromseg
{
   int   fgl;
   long long endpos;
} Chromseg;

/* Why is this a separate struct from Chromseg--seems complicated? --CGG      */
typedef struct _Chromseglist
{
   struct _Chromseg *current;
   struct _Chromseglist *next;
} Chromseglist;

typedef struct _Breaklist
{
    long long addr;
    struct _Breaklist * next;
} Breaklist;

typedef struct _Ibdlist
{
    long long start;
    long long end;
    struct _Ibdlist * next;
} Ibdlist;

typedef struct
{
   int   id;
   int   motherid;
   int   fatherid;
   int   sex;                           /* 0 -> female, 1 -> male.            */
   Chromseglist * maternal;
   Chromseglist * paternal;
} Person;

/* The following structure defines a single gamete array element as read in   */
/* from the input ibd_graph gamete data file.                                 */

typedef struct
{
   char *name;                          /* Name of individual to which this   */
                                        /* gamete belongs.                    */
   int   which;                         /* 0 -> maternal, 1 -> paternal.      */
   int   index;                         /* The subscript for this gamete.     */
   Chromseglist * seglist;              /* Pointer to this gamete's seglist.  */
} Gamete;
typedef Gamete * P_Gamete;

typedef struct _Hashnode
{
   int fgl;
   int index;
   struct _Hashnode * next;
} Hashnode;

/* Data structures used by beaglesim (primarily Chris Glazner's description   */
/* along with subsequent clarifications by Steven Lewis):                     */
/*                                                                            */
/* We have markers and edge nodes.  A slightly confusing thing is that the    */
/* nodes here are actually the edges of the VLMC in, eg, Browning 2006 AJHG.  */
/*                                                                            */
/* markers: []; edges: o                                                      */
/*                                                                            */
/* [] -> o                                                                    */
/* |                                                                          */
/* v                                                                          */
/* [] -> o -> o -> o -> o                                                     */
/* |                                                                          */
/* v                                                                          */
/* [] -> o -> o                                                               */
/*                                                                            */
/* At the previous marker, we have selected a child = I and counted the total */
/* number = N of edges going into that child.  At the next marker, we choose  */
/* a random integer K between 1 and N and traverse the chain of edges until   */
/* the running count of the nodes with parent = I is greater than K.  That    */
/* edge determines the child node at the next marker and the allelic type of  */
/* the allele being simulated.                                                */
/*                                                                            */
/* For a recombination jump (Chris Glazner's name for what Elizabeth wants    */
/* referred to as a linkage disequilibrium relaxation possible jump), we      */
/* follow the same procedure except we can select any edge, not just those    */
/* with a specific parent.                                                    */

typedef struct _marker
{
   struct _edge *edges;
   struct _marker *nextmarker;
   int   index;
} marker;

typedef struct _edge
{
   struct _edge *nextedge;
   int   parent;
   int   child;
   char  allele;
   int   count;
   int   nsamechild;
} edge;

typedef struct
{
   unsigned int z, w, jsr, jcong;
} rngstate;

struct ss_ibd_data_structure
{
   int   ss_size;                       /* Size of the score set.             */
   int   num_ss;                        /* Number of score sets.              */
   int  *index;                         /* Pointer to IBD state index.        */
   int  *number;                        /* Pointer to IBD state number.       */
   int  *label;                         /* Pointer to IBD state label.        */
};
typedef struct ss_ibd_data_structure SS_IBD_DATA_STRUCTURE;

GLOBAL SS_IBD_DATA_STRUCTURE *ss_ibd_data; /* Pointer to a vector of          */
                                  /* score_set_ibd_data structures.           */
GLOBAL Gamete *ibdgraph_gametes;  /* Pointer to a vector of ibd_graph gametes */
                                  /* read from the input ibd_graph gamete     */
                                  /* data file.                               */
GLOBAL int *snp_pos;              /* Vector of SNP locations.                 */
GLOBAL int num_gametes;           /* Number of gametes read from the input    */
                                  /* ibd_graph gamete data file.              */
GLOBAL int largest_fgl;           /* Largest FGL found in the data.           */

/* Prototypes for the subroutines in the IBD_Create files from before they    */
/* were added to Morgan:                                                      */

void new_generation (Person *, Person *, Person **, Person **, int, int, int,
                     long long, double, double);

Chromseglist ** add_chrom_tail (Chromseglist **, int, long long);
Ibdlist ** add_ibd_tail (Ibdlist **, long long, long long);

long long cent2micro (double);
Chromseglist * recombine (Person, Breaklist *);
void weed_chrom (Chromseglist **);
Ibdlist * find_ibd (Chromseglist *, Chromseglist *);

void printbreaks (FILE *, Chromseglist *);
void printibd (FILE *, Ibdlist *, int);
void printgeneration (FILE *, Person *, int, int);

Breaklist * makebreaks (long long, double);
Person * initfounders (int, long long, int, int);
Chromseglist * read_chrom (FILE *);
void free_breaks (Breaklist *);
void free_ibdlist (Ibdlist *);

Hashnode ** add_hash_tail (Hashnode **, int, int);
Chromseglist * get_ibd (Chromseglist *, Chromseglist *);
int  * ibd_nine (int *, int);
void print_table (FILE *, Hashnode **);

/* Prototypes updated for the IBD_Create files after the main subprograms     */
/* were added to Morgan:                                                      */

int  * ibd_fifteen (Gamete, Gamete, Gamete, Gamete, int, int *);
int  populate_snps (const char *, FILE *, const int, const int, char ***);
int  comparGams (const void *, const void *);
void free_generation (Person *, const int);
void free_chromseglist (Chromseglist *);

unsigned int kiss (rngstate *);
double kissunif (rngstate *);
rngstate * initstream (unsigned int);
marker * addmarker (marker *, int);
edge * addedge (edge *, int, int, char, int);

char ** assign_alleles (Hashnode **, char **, int, int);
void output_haplotypes (FILE *, char **, int, int);
void output_genotypes (FILE *, char **, int, int);
void read_gamete_data (FILE *, int *, Gamete **);

void gen_fgl2ibd_labels (int, int, FILE *, FILE *);
int  fgl2ibd_4gam (FILE *, FILE *);

Hashnode ** alloc_hashtable (void);
void fill_hashtable (Hashnode **, int, int *);
int  get_index (int, Hashnode **, int *);

void init_ss_ibd_data (FILE *);
void free_ss_ibd_data (void);
int  fgl_marks (Chromseglist *, int, int *, int *);
void ibd_state (int **, int, int, int *);

void make_names_vector (void);
void free_gamete_data (int, Gamete *);
void free_hashtable (Hashnode **);
void free_marker_list (marker *);
void init_snp_pos (int);

void make_dgl_graphs_file (FILE *, int, Gamete *);
int find_next_change_point (const Chromseglist *, const int, Chromseglist **,
                            int *);

/* Prototypes for the subroutines that were moved to the ibd_create_debug.c   */
/* file.                                                                      */

void testrng (void);
void progbar (int);
void disp_dag_file (marker *);
void prnt_marker_positions (void);
void prnt_gam_segs (const Gamete);
void printWithCommas (int);
char * strrev (char *);

/* Prototypes for the main subprograms:                                       */

int  beaglesim    (int argc, char **argv);
int  beagledag    (int argc, char **argv);
int  simpop_fgl   (int argc, char **argv);
int  simped_fgl   (int argc, char **argv);
int  fgl2haplo    (int argc, char **argv);
int  fgl2ibd      (int argc, char **argv);
int  fgl2dgl      (int argc, char **argv);
int  gunzipit     (char *);

#endif  /*IBD_CREATE_H*/
