/******************************************************************************/
/*                                                                            */
/* Subroutines used to help debug any of the ibd_create.c subprograms.  The   */
/* printWithCommas and strrev subroutines are included because these were     */
/* written for use by other subroutines included.  However, these two         */
/* subroutines maybe should be moved to a file in the Stuff directory.        */
/*                                                                            */
/******************************************************************************/

#include "ibd_create.h"

/******************************************************************************/
/*                                                                            */
/* The testrng subroutine was written by Chris Glazner for testing the random */
/* number generator(s) he wrote inside beaglesim.                             */
/*                                                                            */
/******************************************************************************/

void testrng (void)
{
   rngstate *st1;

   unsigned int k;

   int   i;

   st1 = initstream (1);

   /* Note: these values are taken from the Marsaglia post. The 1,000,257th   */
   /* call to kiss() from starting with this state should give 1372460312.    */
   /* Another 1,000,001 calls should produce the listed subgenerator values.  */

   st1->z = 12345;
   st1->w = 65435;
   st1->jsr = 34221;
   st1->jcong = 12345;

   for (i = 1; i < 1000001 + 256; ++i)
      k = kiss (st1);

   printf (" RNG test: These should be zero: %d ", k - 1372460312U);
   for (i = 1; i < 1000001; ++i)
      kiss (st1);

   printf ("%d %d %d\n", st1->jcong - 1529210297U, st1->jsr - 2642725982U,
           (st1->z << 16) + (st1->w) - 904977562U);
   free (st1);

   return;
}

/******************************************************************************/
/*                                                                            */
/* The progbar (ie, progress bar) subroutine was written by Chris Glazner for */
/* generating a progress bar from within beagelsim.                           */
/*                                                                            */
/******************************************************************************/

void progbar (int percent)
{
   int   prog;
   int   i;

   /* If beaglesim is running as part of the IBD_create gold standards, the   */
   /* progress bar should not be produced because if it is included in the    */
   /* console output file it really messes up the diff command check used in  */
   /* the Makefile.                                     SML - June 27, 2014.  */

   if (DebugLevel == 4096)
      return;

   prog = ceil (percent * 50 / (double) 100);

   printf ("\r [");
   for (i = 0; i < prog; ++i)
      printf ("-");

   for (; i < 50; ++i)
      printf (" ");

   printf ("] %3.d%%", percent);
   fflush (stdout);

   return;
}

/******************************************************************************/
/*                                                                            */
/* The disp_dag_file subroutine can be called to output to stdout a printout  */
/* of the entire DAG file, with rows in reverse order within markers.         */
/*                                                                            */
/******************************************************************************/

void disp_dag_file (marker * mhead)
{
   marker *mcurr;
   edge *ecurr;

   mcurr = mhead;

   while (mcurr != NULL)
   {
      ecurr = mcurr->edges;

      while (ecurr != NULL)
      {
         printf ("%d %d %d %c %d %d\n", mcurr->index, ecurr->parent,
                 ecurr->child, ecurr->allele, ecurr->count, ecurr->nsamechild);

         ecurr = ecurr->nextedge;
      }

      mcurr = mcurr->nextmarker;
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* The prnt_marker_positions subroutine can be called to print out all of the */
/* selected marker numbers along with their corresponding positions in micro  */
/* centimorgans.                                                              */
/*                                                                            */
/******************************************************************************/

void prnt_marker_positions (void)
{
   int   m1;                            /* Current marker subscript.          */

   fprintf (stderr, " Selected marker numbers with their corresponding");
   fprintf (stderr, " positions in micro centimorgans:\n");

   for (m1 = 0; m1 < NMarks; ++m1)
   {
      fprintf (stderr, " (%d@ ", MarSel[m1]);
      printWithCommas (snp_pos[m1]);
      fprintf (stderr, "),");
   }

   fprintf (stderr, " for the %d selected markers.\n\n", NMarks);

   return;
}

/******************************************************************************/
/*                                                                            */
/* The prnt_gam_segs subroutine can be called to print out data for all FGL   */
/* segments for the given gamete.  That is, each FGL followed by its starting */
/* micro centimorgan position, then a '-' followed by its ending micro        */
/* centimorgan position.  The entire range enclosed in parentheses will be    */
/* output.                                                                    */
/*                                                                            */
/******************************************************************************/

void prnt_gam_segs (const Gamete gamete)
{
   Chromseglist *curr_seglist;          /* Current gamete segment list.       */

   char  which[2] = { 'm', 'p' };       /* Character identifying whether the  */
   /*                                      gamete is maternal or paternal.    */

   int   start_pos;                     /* Start position of current segment. */
   int   end_pos;                       /* End position of current segment.   */

   int   numb_segs;                     /* Number of segments in the gamete.  */

   int   curr_fgl;                      /* Current founder genome label.      */

   fprintf (stderr, " FGL segment list for %s's", gamete.name);
   fprintf (stderr, " %caternal gamete", which[gamete.which]);
   fprintf (stderr, " (gamete subscript %d):\n", gamete.index);

   start_pos = 1;
   numb_segs = 0;

   curr_seglist = gamete.seglist;

   while (curr_seglist != NULL)
   {
      curr_fgl = curr_seglist->current->fgl;
      end_pos = curr_seglist->current->endpos;

      fprintf (stderr, " (%d: ", curr_fgl);
      printWithCommas (start_pos);
      fprintf (stderr, "-");
      printWithCommas (end_pos);
      fprintf (stderr, "),");

      curr_seglist = curr_seglist->next;
      start_pos = ++end_pos;
      ++numb_segs;
   }

   fprintf (stderr, " for a total of %d segments.\n\n", numb_segs);

   return;
}

/******************************************************************************/
/*                                                                            */
/* The printWithCommas subroutine can be called to output the input integer   */
/* with commas inserted before every three significant digits.                */
/*                                                                            */
/******************************************************************************/

void printWithCommas (int number)
{
   char  char_version[24];              /* Numbers with at most 5 commas.     */

   int   num_digits;                    /* Number of digits copied so far.    */
   int   char_ndx;                      /* Current subscript in char_version. */

   /* Convert the input number into character form while inserting commas.    */
   /* This loop stores the characters in reverse of the desired output order. */
   /* The order will be reversed by calling the strrev function below.        */

   num_digits = 0;
   char_ndx = 0;

   while (number > 0)
   {
      char_version[char_ndx] = '0' + (number % 10);
      ++num_digits;

      /* Remove the copied digit from the argument integer passed to the      */
      /* the printWithCommas subroutine.                                      */

      number /= 10;
      ++char_ndx;

      if (number > 0 && ((num_digits % 3) == 0))
         char_version[char_ndx++] = ',';
   }
   char_version[char_ndx] = NUL;        /* End the string with a '\0'.        */

   /* Reverse the character string into the order wanted for the output.      */

   strrev (char_version);

   fprintf (stderr, "%s", char_version);

   return;
}

/******************************************************************************/
/*                                                                            */
/* The strrev subroutine can be called to reverse the order of the characters */
/* in the single argument character string.  The character reversal is done   */
/* in place in the string argument.  So that strrev can be called as either a */
/* subroutine or as a function, a pointer to the result is also returned by   */
/* this subroutine.                                                           */
/*                                                                            */
/* The standard C Library does not include a strrev function.  However many   */
/* vendors have added just such a function to their C Libraries.  The         */
/* implementation here is based on one I found during my searches of the      */
/* internet.                                                                  */
/*                                                                            */
/******************************************************************************/

char *strrev (char *str_arg)
{
   char  lead_char;                     /* Holds lead character temporarily.  */

   size_t l_str_arg;                    /* Length of input string in str_arg. */

   int   str_ndx;                       /* Current subscript into str_arg.    */
   int   end_ndx;                       /* Ending subscript into str_arg.     */

   l_str_arg = strlen (str_arg);
   end_ndx = l_str_arg - 1;

   for (str_ndx = 0; str_ndx < end_ndx; ++str_ndx)
   {
      lead_char = str_arg[str_ndx];
      str_arg[str_ndx] = str_arg[end_ndx];
      str_arg[end_ndx] = lead_char;
      --end_ndx;
   }

   return str_arg;
}
