/******************************************************************************/
/*                                                                            */
/* Subroutines called by both simpop_fgl and simped_fgl.                      */
/*                                                                            */
/******************************************************************************/

#include "ibd_create.h"
#include "rans.h"

/******************************************************************************/
/*                                                                            */
/* The cent2micro subroutine converts centimorgans (passed as a double        */
/* precision real number) into micro centimorgans (returned as a long long    */
/* integer).                                                                  */
/*                                                                            */
/******************************************************************************/

long long cent2micro (double input_cent)
{
   long long result;                    /* Chromosome length in micro         */
   /*                                      centimorgans.                      */

   result = (long long) floor ((input_cent * 1E6) + 0.5);

   return result;
}

/******************************************************************************/
/*                                                                            */
/* The recombine subroutine creates a new Chromseglist pointer using both the */
/* maternal and paternal chromosomes belonging to the Person passed as the    */
/* first argument to this subroutine.                                         */
/*                                                                            */
/******************************************************************************/

Chromseglist *recombine (Person x, Breaklist * newbreaks)
{
   Chromseglist **curr_chrom, **tailaddr;
   Chromseglist *curr_first = x.maternal;
   Chromseglist *curr_second = x.paternal;
   Chromseglist *out = NULL;
   Breaklist *curr_break = newbreaks;

   long long segbegin, segend;

   int   which;

   /* start tracking the list tail--begins as a NULL pointer */
   tailaddr = &out;

   segbegin = 0;
   /* randomly choose a chromosome to start on */
   which = rans () < 0.5;

   /* traverse the list of new breaks */
   while (curr_break)
   {
      /* current segment end is the breakpoint we're currently at */
      segend = curr_break->addr;
      /* change curr_chrom's pointee to be the current chromosome */
      curr_chrom = which ? &curr_first : &curr_second;
      /* traverse list until we're past the current segment start */
      while ((*curr_chrom)->current->endpos < segbegin)
      {
         *curr_chrom = (*curr_chrom)->next;
      }
      /* now traverse until we're past the current segment end,
       * but as we go put the breaks into the output chromosome
       * using add_chrom_tail */
      while ((*curr_chrom)->current->endpos < segend)
      {
         tailaddr =
            add_chrom_tail (tailaddr, (*curr_chrom)->current->fgl,
                            (*curr_chrom)->current->endpos);
         *curr_chrom = (*curr_chrom)->next;
      }
      /* finally, put the breakpoint in as a new segment */
      tailaddr = add_chrom_tail (tailaddr, (*curr_chrom)->current->fgl, segend);
      /* set new segment start to previous segment end */
      segbegin = curr_break->addr;

      curr_break = curr_break->next;
      /* switch current chromosome */
      which = (which + 1) % 2;
   }

   /* does the last bit after the last break */
   curr_chrom = which ? &curr_first : &curr_second;
   while ((*curr_chrom)->current->endpos < segbegin)
   {
      *curr_chrom = (*curr_chrom)->next;
   }
   while (*curr_chrom)
   {
      tailaddr =
         add_chrom_tail (tailaddr, (*curr_chrom)->current->fgl,
                         (*curr_chrom)->current->endpos);
      *curr_chrom = (*curr_chrom)->next;
   }
   /* eliminates redundant nodes in the output chromosome */
   weed_chrom (&out);

   return out;
}

Chromseglist **add_chrom_tail (Chromseglist ** tailaddr, int newfgl,
                               long long newendpos)
{
   /* Takes a Chromseglist pointer by reference, creates a new node, and      */
   /* changes the passed pointer to point to the new node.                    */

   Chromseglist *out;

   out = (Chromseglist *) malloc (sizeof (Chromseglist));
   alloc_chk (WHERE, out);
   out->current = (Chromseg *) malloc (sizeof (Chromseg));
   alloc_chk (WHERE, out->current);

   out->current->fgl = newfgl;
   out->current->endpos = newendpos;
   out->next = NULL;

   *tailaddr = out;

   return &((*tailaddr)->next);
}

void weed_chrom (Chromseglist ** x)
/* Remove redundant junctions in the chromosome which is passed by reference. */
/* This makes things cleaner and makes finding IBD easier.                    */
{
   Chromseglist *curr, *last;
   Chromseglist **curr_ptr;

   curr = (*x)->next;
   last = *x;
   curr_ptr = x;

   while (curr)
   {
      /* if fgl is the same, cut that node out of the list */
      if (curr->current->fgl == last->current->fgl)
      {
         (*curr_ptr) = curr;
         free (last->current);
         free (last);
      }
      else
         /* otherwise, advance the (pointer to pointer) */
         curr_ptr = &last->next;
      last = curr;
      curr = curr->next;
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* The makebreaks subroutine is called to generate a linked list of change    */
/* points along a chromosome with length passed as the first argument (in     */
/* micro centimorgans).                                                       */
/*                                                                            */
/* The rate argument (the second argument) is a double precision number,      */
/* expressing expected number of breaks per chrlength (the first argument).   */
/*                                                                            */
/******************************************************************************/

Breaklist *makebreaks (long long chrlength, double rate)
{
   Breaklist *current, *head = NULL;

   long long nextbreak;
   long long exprand;
   long long runsum;

   double rexp_var;

   runsum = 0;

   while (TRUE)
   {
      /* The linked list of chromosome segments is being contructed backwards */
      /* by keeping track of the head of the list.                            */

      /* Draw a random variate from the exponential distribution with the     */
      /* given rate.                                                          */

      rexp_var = rexp (rate);

      /* This would get wacky if chrlength were unsigned.                     */
      exprand = (long long) floor (chrlength * rexp_var);
      nextbreak = max (exprand, 1);

      runsum += nextbreak;
      if (runsum > chrlength)
         break;

      current = (Breaklist *) malloc (sizeof (Breaklist));
      alloc_chk (WHERE, current);
      current->addr = chrlength - runsum;
      current->next = head;
      head = current;
   }

   return head;
}

void free_breaks (Breaklist * list)
{
   Breaklist *curr, *prev;

   if (!list)
      return;
   prev = list;
   curr = list->next;
   while (curr)
   {
      free (prev);
      prev = curr;
      curr = curr->next;
   }
   free (prev);

   return;
}

/* Id number included just to tag output for later use.                       */
void printbreaks (FILE * oScorF, Chromseglist * x)
{
   Chromseglist *curr = x;

   while (curr)
   {
      fprintf (oScorF, " %d %llu", curr->current->fgl, curr->current->endpos);
      curr = curr->next;
   }
   fprintf (oScorF, "\n");

   return;
}

/******************************************************************************/
/*                                                                            */
/* The free_generation subroutine frees up the memory allocated for a single  */
/* generation of individuals.  The first argument passed to free_generation   */
/* is a Person pointer to the origin of the generation vector.  The second    */
/* argument is an integer providing the number of individuals allocated for   */
/* the generation vector.                                                     */
/*                                                                            */
/******************************************************************************/

void free_generation (Person * individuals, const int num_indivs)
{
   int   i1;                            /* Subscript to current individual.   */

   for (i1 = 0; i1 < num_indivs; ++i1)
   {
      free_chromseglist (individuals[i1].maternal);
      free_chromseglist (individuals[i1].paternal);
   }

   free_1d ((void *) &individuals);

   return;
}

/******************************************************************************/
/*                                                                            */
/* The free_chromseglist subroutine frees up the memory allocated for the     */
/* linked list of chromosome segments pointed to by the one argument to the   */
/* subroutine.                                                                */
/*                                                                            */
/******************************************************************************/

void free_chromseglist (Chromseglist * first_chromseg)
{
   Chromseglist *p_curr_seg;            /* Pointer to the current chromosome  */
   /*                                      segment element.                   */
   Chromseglist *p_next_seg;            /* Pointer to the next chromosome     */
   /*                                      segment element.                   */

   p_curr_seg = first_chromseg;
   while (p_curr_seg != NULL)
   {
      free (p_curr_seg->current);
      p_next_seg = p_curr_seg->next;
      free (p_curr_seg);
      p_curr_seg = p_next_seg;
   }

   return;
}
