/******************************************************************************/
/*                                                                            */
/* The haplo_debug_phased subroutine calls the print_one_line subroutine to   */
/* print each line of debugging information to the oFile.                     */
/*                                                                            */
/******************************************************************************/

void print_one_line (char *curr_ibd_state, double ibd_total, int first_state,
                     int ns_in_orbit, double **ibd_final, int m1, FILE * oFile)
{
   int   s1;                            /* Subscript to current state within  */
   /*                                      the ibd_final[][] matrix.          */

   fprintf (oFile, " %4s", curr_ibd_state);
   fprintf (oFile, " %7.4f  ", ibd_total);

   s1 = first_state;
   while (ns_in_orbit > 0)
   {
      fprintf (oFile, " %7.4f", ibd_final[m1][s1]);
      --ns_in_orbit;
      ++s1;
   }

   fprintf (oFile, "\n");

   return;
}

/******************************************************************************/
/*                                                                            */
/* The haplo_debug_phased subroutine totals the gene IBD probabilities within */
/* each orbit, then prints this total probability and each IBD state          */
/* probability to the file selected as the fourth argument.                   */
/*                                                                            */
/******************************************************************************/

void haplo_debug_phased (int mark_width, int num_states, double **ibd_final,
                         FILE * oFile)
{
   char *curr_ibd_state = NULL;
   char *next_ibd_state = NULL;

   double ibd_total;                    /* Total probability in current IBD   */
   /*                                      orbit.                             */

   int   ns_in_orbit;                   /* Number of states in current orbit. */

   int   state_ndx;                     /* Subscript to input IBD gene state. */
   int   first_state;                   /* Subscript to first state in orbit. */

   int   first_line;                    /* Set to TRUE if this is the first   */
   /*                                      line of output for current marker. */
   int   m1;                            /* Subscript to current marker.       */

   for (m1 = 0; m1 < NMarks; ++m1)      /* For each marker,                   */
   {
      fprintf (oFile, " [%*d] ", mark_width, MarSel[m1]);
      first_line = TRUE;

      curr_ibd_state = compute_ibd_state (0, 4);
      first_state = 0;
      ibd_total = 0.0;
      ns_in_orbit = 0;

      /* For each possible IBD state,                                         */

      for (state_ndx = 0; state_ndx < num_states; ++state_ndx)
      {
         /* Construct the possible IBD state label.                           */

         free_1d ((void *) &next_ibd_state);
         next_ibd_state = compute_ibd_state (state_ndx, 4);

         if (strncmp (curr_ibd_state, next_ibd_state, 4) != 0)
         {
            if (first_line == FALSE)
               fprintf (oFile, "  %*s  ", mark_width, " ");

            print_one_line (curr_ibd_state, ibd_total, first_state, ns_in_orbit,
                            ibd_final, m1, oFile);
            first_line = FALSE;

            strcpy (curr_ibd_state, next_ibd_state);
            first_state = state_ndx;
            ibd_total = 0.0;
            ns_in_orbit = 0;
         }

         ibd_total += ibd_final[m1][state_ndx];
         ++ns_in_orbit;
      }

      /* Print the line of debugging output for the last gene IBD orbit.      */

      fprintf (oFile, "  %*s  ", mark_width, " ");
      print_one_line (curr_ibd_state, ibd_total, first_state, ns_in_orbit,
                      ibd_final, m1, oFile);
      fprintf (oFile, "\n");

      free_1d ((void *) &curr_ibd_state);
      free_1d ((void *) &next_ibd_state);
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* The haplo_debug_unphased subroutine prints for each unphased reduced state */
/* the gene IBD state label followed by the gene IBD probability to the file  */
/* selected as the fifth argument.                                            */
/*                                                                            */
/******************************************************************************/

void haplo_debug_unphased (int mark_width, int ns_final, double **ibd_final,
                           char **state_ids, FILE * oFile)
{
   int   maxlen_state_id;               /* Maximum length of any state_id.    */

   int   state_ndx;                     /* Subscript to input IBD gene state. */
   int   m1;                            /* Subscript to current marker.       */

   /* Find out how long the longest state_id is.                              */
   maxlen_state_id = 0;
   for (state_ndx = 0; state_ndx < ns_final; ++state_ndx)
      maxlen_state_id = max (maxlen_state_id, strlen (state_ids[state_ndx]));

   for (m1 = 0; m1 < NMarks; ++m1)      /* For each marker,                   */
   {
      fprintf (oFile, " [%*d] ", mark_width, MarSel[m1]);

      /* For each unphased reduced possible IBD state,                        */
      for (state_ndx = 0; state_ndx < ns_final; ++state_ndx)
      {
         if (state_ndx > 0)
            fprintf (oFile, "  %*s  ", mark_width, " ");

         fprintf (oFile, " %-*s", maxlen_state_id, state_ids[state_ndx]);
         fprintf (oFile, " %7.4f\n", ibd_final[m1][state_ndx]);
      }

      fprintf (oFile, "\n");
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* Debugging subroutine called to output the qq_ibd matrix returned from the  */
/* hap_bck_compute subroutine to the stderr file.  This subroutine has been   */
/* replaced by the haplo_debug_qq_ibd subroutine.                             */
/*                                                                            */
/******************************************************************************/

void haplo_debug_initial_stderr (int mark_width, double **qq_ibd)
{
   int   mark_ndx;                      /* Subscript to current marker.       */
   int   state_ndx;                     /* Subscript to input IBD gene state. */

   for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
   {
      fprintf (stderr, " [%*d]", mark_width, MarSel[mark_ndx]);
      for (state_ndx = 0; state_ndx < num_states; ++state_ndx)
      {
         fprintf (stderr, " %7.4f", qq_ibd[mark_ndx][state_ndx]);
      }

      fprintf (stderr, "\n");
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* Debugging subroutine called to output the qq_ibd matrix returned from the  */
/* hap_bck_compute subroutine to an oFile.  This subroutine replaces the      */
/* haplo_debug_initial_stderr subroutine.                                     */
/*                                                                            */
/******************************************************************************/

void haplo_debug_qq_ibd (int mark_width, int num_curr_states, double **qq_ibd,
                         FILE * oFile)
{
   int   mark_ndx;                      /* Subscript to current marker.       */
   int   state_ndx;                     /* Subscript to input IBD gene state. */

   for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
   {
      fprintf (oFile, " [%*d]", mark_width, MarSel[mark_ndx]);
      for (state_ndx = 0; state_ndx < num_curr_states; ++state_ndx)
      {
         fprintf (oFile, " %7.4f", qq_ibd[mark_ndx][state_ndx]);
      }
      fprintf (oFile, "\n");
   }

   fprintf (oFile, "\n");

   return;
}

/******************************************************************************/
/*                                                                            */
/* Debugging subroutine called to output the phase data as read in from the   */
/* user's phasing data file.                                                  */
/*                                                                            */
/* The arguments to this subroutine were declared in the original calling     */
/* subroutines as:                                                            */
/*                                                                            */
/*        int s1;                          Subscript to score set.            */
/* static int nc;                          Character length of longest name.  */
/*                                                                            */
/******************************************************************************/

void print_phase_data (int s1, int nc)
{
   int   m1;                            /* Subscript to current marker.       */

   printf (" Individual name %*s:", nc, ObsNam[2 * s1]);
   for (m1 = 0; m1 < NMarks; ++m1)
      printf (" %d", phasdatindiv[2 * s1][m1]);
   printf ("\n");

   printf (" Individual name %*s:", nc, ObsNam[2 * s1 + 1]);
   for (m1 = 0; m1 < NMarks; ++m1)
      printf (" %d", phasdatindiv[2 * s1 + 1][m1]);
   printf ("\n");

   printf (" Pair (4-haplotypes) %*s %*s:", nc, ObsNam[2 * s1], nc,
           ObsNam[2 * s1 + 1]);
   for (m1 = 0; m1 < NMarks; ++m1)
      printf (" %d", phasdat[m1]);
   printf ("\n");

   fflush (stdout);

   return;
}
