/******************************************************************************/
/*                                                                            */
/* The unique_int subroutine creates a vector of integers consisting of the   */
/* integers passed in the second argument (input), with no duplicate values.  */
/* The values will be in the same order as in the input argument except that  */
/* any repeated values will be deleted.                                       */
/*                                                                            */
/******************************************************************************/

void unique_int (int num_input, int *input, int *num_result, int **result)
{
   int  *duplicated;                    /* A vector of 0's and 1's reflecting */
   /*                                      where there are duplicates.        */

   int   inp_ndx;                       /* Subscript in the input vector.     */
   int   comp_ndx;                      /* Subscript to comparison integer.   */
   int   uniq_ndx;                      /* Subscript in the unique vector.    */

   /* Create a one dimensional vector of integers set to 0 or 1, where 0      */
   /* means the corresponding entry in input is not duplicated and 1 means    */
   /* the corresponding entry in input is a duplicate of a previous entry.    */

   duplicated = alloc_int1d (num_input);

   for (inp_ndx = 1; inp_ndx < num_input; ++inp_ndx)
   {
      for (comp_ndx = 0; comp_ndx < inp_ndx; ++comp_ndx)
      {
         if (input[inp_ndx] == input[comp_ndx])
         {
            duplicated[inp_ndx] = TRUE;
            break;
         }
      }
   }

   /* Allocate enough storage to create a vector of unique integers.          */

   *result = alloc_int1d (num_input);

   uniq_ndx = 0;
   for (inp_ndx = 0; inp_ndx < num_input; ++inp_ndx)
   {
      if (duplicated[inp_ndx] == FALSE)
      {
         (*result)[uniq_ndx] = input[inp_ndx];
         ++uniq_ndx;
      }
   }

   free (duplicated);

   *num_result = uniq_ndx;

   return;
}

/******************************************************************************/
/*                                                                            */
/* Subroutine to turn an input IBD state vector to use an equivalent minimum  */
/* numbering.  For example, if the input is (2, 3, 1, 1), the equivalent      */
/* minimum numbering output would be (1, 2, 3, 3).                            */
/*                                                                            */
/* The first argument is the number of gametes making up the input IBD        */
/* state vector.  The next argument is the input integer vector representing  */
/* an IBD state.  The third argument is a pointer to an integer vector to be  */
/* used by the rplc subroutine to return the equivalent minimum numbering IBD */
/* state vector in.                                                           */
/*                                                                            */
/******************************************************************************/

void rplc (int num_gams, int *input_state, int *output)
{
   int   num_uniq_gams;                 /* Number of unique gametes.          */
   int  *unique_gams;                   /* Vector of unique gametes.          */

   int   uniq_ndx;                      /* Subscript in unique state vector.  */
   int   out_ndx;                       /* Subscript to output state vector.  */

   /* Call the unique_int subroutine to create a vector of the unique values  */
   /* present in the input integer vector, input_state.                       */

   unique_int (num_gams, input_state, &num_uniq_gams, &unique_gams);

   /* This ensures that when going from state to label, the label is always   */
   /* the first label representing that state.                                */

   for (uniq_ndx = 0; uniq_ndx < num_uniq_gams; ++uniq_ndx)
   {
      for (out_ndx = 0; out_ndx < num_gams; ++out_ndx)
      {
         if (input_state[out_ndx] == unique_gams[uniq_ndx])
            output[out_ndx] = uniq_ndx + 1;
      }
   }

   free (unique_gams);

   return;
}

/******************************************************************************/
/*                                                                            */
/* The get_ibd_state_mat subroutine is called to allocate and initialize a    */
/* matrix of IBD states.  The number of gametes in each IBD state is the      */
/* first argument.  The get_ibd_state_mat subroutine returns the number of    */
/* unique IBD states as its second argument.  The third argument is a vector  */
/* of 0's and 1's reflecting where there are duplicate rows, returned by the  */
/* get_ibd_state_mat subroutine so that the set_tmat_2011 subroutine can use  */
/* it.  The fourth argument is the lexigraphically ordered matrix of IBD      */
/* states with all the duplicated states removed.                             */
/*                                                                            */
/******************************************************************************/

void get_ibd_state_mat (int num_gams, int *num_uniq_states, int **duplic_vector,
                        int ***state_mat)
{
   size_t row_length;                   /* Length of a row in any of the smat */
   /*                                      matrices, in number of bytes.      */

   int   max_label;                     /* Maximum label for the given number */
   /*                                      of gametes.                        */

   int **smat;                          /* The IBD state matrix in the non-   */
   /*                                      minimal integer notation.          */

   int **smat1;                         /* The IBD state matrix in minimal    */
   /*                                      integer notation.                  */

   int **smat2;                         /* The IBD state matrix with all      */
   /*                                      duplicated rows removed.           */

   int   num_uniq_rows;                 /* Number of unique rows in smat1.    */

   int   inp_ndx;                       /* Subscript in the smat1 matrix.     */
   int   comp_ndx;                      /* Subscript to comparison row.       */

   int   i1, i2;                        /* Subscripts to use in for loops.    */

/******************************************************************************/
/*                                                                            */
/* The start of the executable part of the subroutine is here.                */
/*                                                                            */
/******************************************************************************/

   max_label = CDgg[num_gams];
   smat = alloc_int2d (max_label + 1, num_gams);

   /* Create a matrix containing all possible IBD states.  Note that there    */
   /* can be more than one label generating the same IBD state.               */

   for (i1 = 0; i1 <= max_label; ++i1)
      invert_label (i1, num_gams, smat[i1]);

   /* Change the numbering of the IBD states to standard numbering.           */

   smat1 = alloc_int2d (max_label + 1, num_gams);

   for (i1 = 0; i1 <= max_label; ++i1)
      rplc (num_gams, smat[i1], smat1[i1]);

   free_2d ((void *) &smat);

   if (DebugLevel == 1)
   {
      for (i1 = 0; i1 <= max_label; ++i1)
      {
         printf ("For label = %3d, ", i1);
         printf ("smat1 is ");
         for (i2 = 0; i2 < num_gams; ++i2)
            printf ("%2d", smat1[i1][i2]);
         printf ("\n");
      }
      printf ("\n");
   }

   /* Create a one dimensional vector of integers set to 0 or 1, where 0      */
   /* means the corresponding row in smat1 is not duplicated and 1 means the  */
   /* corresponding row in smat1 is a duplicate of a previous row.  Note that */
   /* the alloc_int1d subroutine initializes duplicated_rows to all 0's.      */

   /* Initially assume there are (max_label + 1) unique rows in smat1.        */

   num_uniq_rows = max_label + 1;
   duplicated_rows = alloc_int1d (num_uniq_rows);

   row_length = sizeof (int) * num_gams;

   for (inp_ndx = 1; inp_ndx <= max_label; ++inp_ndx)
   {
      for (comp_ndx = 0; comp_ndx < inp_ndx; ++comp_ndx)
      {
         if (memcmp (smat1[inp_ndx], smat1[comp_ndx], row_length) == 0)
         {
            duplicated_rows[inp_ndx] = TRUE;
            --num_uniq_rows;            /* Decrement count of unique rows.    */
            break;
         }
      }
   }

/******************************************************************************/
/*                                                                            */
/* Reduce a state matrix to the unique rows only, removing duplicate rows.    */
/*                                                                            */
/******************************************************************************/

   /* Allocate enough storage to create a matrix of unique rows.              */

   smat2 = alloc_int2d (num_uniq_rows, num_gams);

   for (i1 = 0, i2 = 0; i1 <= max_label; ++i1)
   {
      if (duplicated_rows[i1] == FALSE)
      {
         memcpy (smat2[i2], smat1[i1], row_length);
         ++i2;
      }
   }

   if (DebugLevel == 1)
   {
      for (i1 = 0; i1 < num_uniq_rows; ++i1)
      {
         printf ("For   row = %3d, ", i1 + 1);
         printf ("smat2 is ");
         for (i2 = 0; i2 < num_gams; ++i2)
            printf ("%2d", smat2[i1][i2]);
         printf ("\n");
      }
      printf ("\n");
   }

   free_2d ((void *) &smat1);

   *num_uniq_states = num_uniq_rows;
   *duplic_vector = duplicated_rows;
   *state_mat = smat2;

   return;
}

/******************************************************************************/
/*                                                                            */
/* Given gametes (e.g. gametes=c(1,2,3,3)) and beta (the pairwise kinship     */
/* probability), this subroutine returns a matrix with 4 columns: the labels  */
/* of all the possible new IBD states, the frequency of transitions to each   */
/* IBD state, the rate of getting to that IBD state, and the index (in        */
/* transitions' t2) to identify the IBD state's leaving gamete.               */
/*                                                                            */
/* Note that some states will be redundant (more than one label per state),   */
/* but this subroutine will calculate transitions to only the label that is   */
/* the first representation of that state.                                    */
/*                                                                            */
/* The arguments to this subroutine are:                                      */
/*   num_gams = number of gametes in the second argument.                     */
/*   gametes = integer vector of gametes representing the IBD state we are    */
/*     coming from.                                                           */
/*   beta = the pairwise kinship probability, which determines the scaled     */
/*     probability of joining existing ibd groups.                            */
/*                                                                            */
/******************************************************************************/

double **transitions (int num_gams, int *gametes, double beta)
{
   double **t5;                         /* Four column matrix to be returned  */
   /*                                      by the transitions subroutine.     */

   double *base_rates;                  /* A vector of transition rates.      */
   double *all_rates;                   /* A vector where the base_rates are  */
   /*                                 repeated for each potential new gamete. */

   double tot_rate;                     /* Computed total transition rate.    */

   int  *gps;                           /* Unique gametes in the input.       */

   int  *gpsz;                          /* Table allocated for counting the   */
   /*                                    number of occurences of each gamete. */

   int  *newguy;                        /* A vector of potential gametes.     */

   int **t2;                            /* A matrix consisting of all the     */
   /*                            potential IBD states we could transition to. */

   int **t3;                            /* Matrix of all potential IBD states */
   /*                                      reduced to standard IBD notation.  */

   int  *t4;                            /* Vector of the labels corresponding */
   /*                                      to each IBD state present in t3.   */

   int  *s2;                            /* Vector of the indices in t2 to     */
   /*                                      identify each leaving gamete.      */

   int   gam_ndx;                       /* Subscript in gametes.              */
   int   gps_ndx;                       /* Subscript in gps.                  */
   int   row_ndx;                       /* Row subscript.                     */

   int   num_uniq_gams;                 /* Number of unique gametes.          */
   int   num_base_rates;                /* Number of entries in base_rates.   */
   int   num_all_rates;                 /* Number of all possible rates.      */

   int   offset;                        /* Current offset into t2.            */

   int   label;                         /* IBD state vector label.            */
   int   num_transitions;               /* Count of number of transitions.    */

   int   max_label;                     /* Maximum label for the given number */
   /*                                      of gametes in the IBD state.       */

   int   i1, i2, i3;                    /* Subscripts to use in for loops.    */

/******************************************************************************/
/*                                                                            */
/* The start of the executable part of the subroutine is here.                */
/*                                                                            */
/******************************************************************************/

   /* The unique_int subroutine allocates storage for gps.                    */

   unique_int (num_gams, gametes, &num_uniq_gams, &gps);

   /* Allocate storage for the vector of counts of how many gametes are in    */
   /* each group.  The alloc_int1d subroutine initializes all counts to 0.    */

   gpsz = alloc_int1d (num_uniq_gams);

   for (gam_ndx = 0; gam_ndx < num_gams; ++gam_ndx)
   {
      for (gps_ndx = 0; gps_ndx < num_uniq_gams; ++gps_ndx)
      {
         if (gametes[gam_ndx] == gps[gps_ndx])
         {
            ++gpsz[gps_ndx];
            break;
         }
      }
   }

   num_base_rates = num_uniq_gams + 1;
   newguy = alloc_int1d (num_base_rates);
   base_rates = alloc_dbl1d (num_base_rates);

   /* Copy the unique gamete numbers from gps to the first num_uniq_gams      */
   /* elements of newguy.  The last element of newguy needs to be set to a    */
   /* gamete number not already in use, such as num_gams + 1.                 */

   for (i1 = 0; i1 < num_uniq_gams; ++i1)
      newguy[i1] = gps[i1];
   newguy[num_uniq_gams] = num_gams + 1;

   /* Initialize the vector of base_rates.                                    */

   for (i1 = 0; i1 < num_uniq_gams; ++i1)
      base_rates[i1] = gpsz[i1] * beta;
   base_rates[num_uniq_gams] = 1.0 - beta;

   /* Calculate all potential transitions and their respective rates.         */

   num_all_rates = (num_gams + 1) * num_base_rates;
   t2 = alloc_int2d (num_all_rates, num_gams);
   s2 = alloc_int1d (num_all_rates);
   all_rates = alloc_dbl1d (num_all_rates);

   /* Set the indexing in s2 = (12345601234560 ......).                       */

   i3 = 0;
   for (i1 = 0; i1 < num_base_rates; ++i1)
   {
      for (i2 = 0; i2 < num_gams; ++i2)
      {
         s2[i3] = i2 + 1;
         ++i3;
      }
      s2[i3] = 0;
      ++i3;
   }

   /* For each arrival.                                                       */
   for (i1 = 0; i1 < num_base_rates; ++i1)
   {
      offset = (num_gams + 1) * i1;

      /* For each departure of an existing gamete.                            */
      for (i2 = 0; i2 <= num_gams; ++i2)
      {
         row_ndx = offset + i2;

         /* Copy beginning state gametes to t2.                               */
         for (i3 = 0; i3 < num_gams; ++i3)
            t2[row_ndx][i3] = gametes[i3];

         /* Set the rate of each potential transition.                        */
         all_rates[row_ndx] = base_rates[i1];

         /* The last row is for when the new guy leaves right away.           */
         if (i2 < num_gams)
            t2[row_ndx][i2] = newguy[i1];
      }
   }

   /* Change to standard IBD notation.                                        */

   t3 = alloc_int2d (num_all_rates, num_gams);

   for (i1 = 0; i1 < num_all_rates; ++i1)
      rplc (num_gams, t2[i1], t3[i1]);

   /* Calculate IBD labels corresponding to each potential state in t3.       */

   t4 = alloc_int1d (num_all_rates);

   for (i1 = 0; i1 < num_all_rates; ++i1)
   {
      label = 0;
      for (i2 = 0; i2 < num_gams; ++i2)
         label = (label * i2) + (t3[i1][i2] - 1);
      t4[i1] = label;
   }

   /* Compute the transition frequencies and rates.                           */

   max_label = CDgg[num_gams];
   t5 = alloc_dbl2d (max_label + 1, 4);

   for (i1 = 0; i1 <= max_label; ++i1)
   {
      num_transitions = 0;
      tot_rate = 0.0;

      /* In the following for loop i2 is the subscript into s2 of the current */
      /* i1'th state and s2[i2] is the leaving gamete.                        */
      for (i2 = 0; i2 < num_all_rates; ++i2)
      {
         if (t4[i2] == i1)
         {
            ++num_transitions;
            tot_rate += all_rates[i2];
            t5[i1][3] = s2[i2];
         }
      }

      t5[i1][0] = i1;
      t5[i1][1] = num_transitions;
      t5[i1][2] = tot_rate;
   }

   /* Free up the storage allocated in the transitions subroutine, other than */
   /* the storage allocated for the t5 matrix.                                */

   free_1d ((void *) &gps);
   free_1d ((void *) &gpsz);
   free_1d ((void *) &newguy);

   free_1d ((void *) &base_rates);
   free_1d ((void *) &all_rates);
   free_1d ((void *) &s2);

   free_2d ((void *) &t2);
   free_2d ((void *) &t3);
   free_1d ((void *) &t4);

   return (t5);
}

/******************************************************************************/
/*                                                                            */
/* Call the set_tmat_2011 subroutine to create a transition matrix from the   */
/* number of gametes and the beta probability.  The number of gametes         */
/* in each IBD state is the first argument.  The second argument is the       */
/* kinship probability (beta) which determines that a proposed new gamete     */
/* will join each existing ibd group.  The third argument is the              */
/* lexigraphically ordered transition matrix with all the duplicated states   */
/* removed.                                                                   */
/*                                                                            */
/******************************************************************************/

void set_tmat_2011 (int num_gams, double beta, double ***tmat_ret)
{
   double **redtmat;                    /* Local copy of transition matrix.   */
   double **redgone;                    /* Local matrix specifying leaving    */
   /*                                      gamete.                            */

   double **t5;                         /* Matrix returned by transitions (). */

   double *row_sums;                    /* Sums of the rows of the transition */
   /*                                      matrix excluding the diagonal.     */

   int   max_label;                     /* Maximum label for the given number */
   /*                                      of gametes.                        */

   int   i1, i2, i3;                    /* Subscripts to use in for loops.    */

/******************************************************************************/
/*                                                                            */
/* Create redtmat.  This version of the transition matrix will not include    */
/* any IBD state which cannot be transitioned to.  In other words, there will */
/* be no all zeroes columns in this matrix.                                   */
/*                                                                            */
/* Also create redgone.  This matrix is used to retain the index of the       */
/* leaving gamete.                                                            */
/*                                                                            */
/******************************************************************************/

   max_label = CDgg[num_gams];
   redtmat = alloc_dbl2d (num_states, num_states);
   redgone = alloc_dbl2d (num_states, num_states);

   for (i1 = 0; i1 < num_states; ++i1)
   {
      t5 = transitions (num_gams, state_matrix[i1], beta);

      for (i2 = 0, i3 = 0; i2 <= max_label; ++i2)
      {
         if (duplicated_rows[i2] == FALSE)
         {
            redtmat[i1][i3] = t5[i2][2];
            redgone[i1][i3] = t5[i2][3];
            ++i3;
         }
      }

      free_2d ((void *) &t5);
   }

   if (DebugLevel == 11)
   {
      printf (" redtmat is:\n");
      for (i1 = 0; i1 < num_states; ++i1)
      {
         printf (" row %3d =  ", i1 + 1);
         for (i2 = 0; i2 < num_states; ++i2)
            printf (" %5.2f", redtmat[i1][i2]);
         printf ("\n");
      }
      printf ("\n");

      printf (" redgone is:\n");
      for (i1 = 0; i1 < num_states; ++i1)
      {
         printf (" row %3d =  ", i1 + 1);
         for (i2 = 0; i2 < num_states; ++i2)
            printf (" %5.2f", redgone[i1][i2]);
         printf ("\n");
      }
      printf ("\n\n");
   }

/******************************************************************************/
/*                                                                            */
/* The diagonal elements of redtmat need to be replaced by minus the sum of   */
/* the other elements in each row of redtmat.  This will produce a valid      */
/* transition rate matrix to be returned to the caller of set_tmat_2011.      */
/* Note that alloc_dbl1d initializes the elements of row_sums to all zeroes.  */
/*                                                                            */
/******************************************************************************/

   row_sums = alloc_dbl1d (num_states);

   for (i1 = 0; i1 < num_states; ++i1)
   {
      for (i2 = 0; i2 < num_states; ++i2)
      {
         if (i1 != i2)                  /* If not a diagonal element.         */
            row_sums[i1] += redtmat[i1][i2];
      }
   }

   /* Replace the diagonal elements of redtmat with the negative of the       */
   /* corresponding element of row_sums.                                      */

   for (i1 = 0; i1 < num_states; ++i1)
   {
      redtmat[i1][i1] = 0.0 - row_sums[i1];
   }

   free_2d ((void *) &redgone);
   free_1d ((void *) &row_sums);
   *tmat_ret = redtmat;

   return;
}
