/******************************************************************************/
/*                                                                            */
/* FUNCTION DEFINITIONS                                                       */
/*                                                                            */
/* Assume IBDGraph_ContainsNode (IBDGraph * g, int node).                     */
/* Step through until the node is contained in the "active" graph,            */
/* then start a new graph.                                                    */
/*                                                                            */
/******************************************************************************/

#include "errsprogs.h"
#include "stuff.h"
#include "IBDgraph.h"

static inline bool verifyKeyAndAdvanceStream (char *name, FILE * fp)
{
   int   pos, r;

   do
   {
      r = fgetc (fp);

      if (unlikely (r == EOF))
         return false;
   }
   while (isspace (r));

   if (r != name[0])
      return false;

   for (pos = 1; pos < EDGE_NAME_BUFFER_SIZE; ++pos)
   {
      r = fgetc (fp);

      if (unlikely (r == EOF))
         return false;

      if (isspace (r))
         return true;

      if (r != name[pos])
         return false;
   }

   do
   {
      r = fgetc (fp);

      if (unlikely (r == EOF))
         return false;
   }
   while (!isspace (r));

   return true;
}

static inline bool getKeyFromFileStream (HashKey * key, char *name, FILE * fp)
{
   /* This version will handle anything!                                      */
   Hk_CLEAR (key);

   bool  starting = true, first = true;
   bool  done, is_int;

   char  buffer[EDGE_NAME_BUFFER_SIZE];

   long  n;
   int   pos;
   int   r;

   /* First get rid of blank characters.                                      */
   while (true)
   {
      done = false;
      is_int = true;

      for (pos = 0; pos < EDGE_NAME_BUFFER_SIZE; ++pos)
      {
         do
         {
            r = fgetc (fp);

            if (unlikely (r == EOF))
               return false;
         }
         while (unlikely (starting && isspace (r)));

         if (isspace (r))
         {
            buffer[pos] = '\0';
            done = true;
            if (first && name != NULL)
               name[pos] = '\0';
            break;
         }
         else
         {
            buffer[pos] = (char) r;
            starting = false;
            if (first && name != NULL)
               name[pos] = buffer[pos];

            is_int &= ! !isdigit (r);
         }
      }

      /* Detect case of exactly a 64 character string; this is next.          */
      if (unlikely (pos == 0))
         return true;

      if (likely (Hk_ISZERO (key)))
      {
         if (likely (done))
         {
            /* See if we can use the cheaper integer hash function            */
            if (is_int)
            {
               n = atol (buffer);

               if (likely (n != 0))
               {
                  *key = Hk_FromInt (n);
                  return true;
               }
            }

            *key = Hk_FromCharBuffer (buffer, pos);
            return true;
         }
         else
         {
            assert (pos == EDGE_NAME_BUFFER_SIZE);
            *key = Hk_FromCharBuffer (buffer, EDGE_NAME_BUFFER_SIZE);
         }
      }
      else
      {
         HashKey new_key = Hk_FromCharBuffer (buffer, pos);
         Hk_InplaceCombine (key, &new_key);

         if (done)
            return true;
      }
   }
}

void createIBDGraphs (char *file, IBDGraphList * ibd_graphs)
{
   FILE *fp;

   IBDGraph *ibd_graph;
   IBDGraphEdge *e;
   IBDGraphNode *cur_node;
   HashKey key;

   char  edge[EDGE_NAME_BUFFER_SIZE];

   int   i, k;
   int   ibd0;
   int   ibd1;
   int   changes;
   int   ibd;
   int   change_pos;
   int   cur_range_min = 0;
   int   mr_graph_min = 0;

   /* Using new functions IBDGraphContainsEdge and IBDGraphContainsNode to    */
   /* build graphs by starting a new graph when a node is repeated.           */

   /* Now go through NUM_GRAPHS of ibd graphs and build them.                 */

   /* Opening file.                                                           */
   fp = fopen (file, "r");

   if (fp == NULL)
   {
      fprintf (stderr, "\nError!  File %s not found.  Aborting.\n\n", file);
      exit (1);
   }

   /* printf ("Preparing graphs.\n\n"); */

   /* Creating IBD graphs until file is empty.                                */

   /* i keeps track of the number of graphs generated and                     */
   /* k keeps the while loop going until reach EOF.                           */

   i = 1;
   k = 1;

   /* Since this is the first run through, there are no edges in the graph.   */
   ibd_graph = NewIBDGraph (i);

   /* printf ("\nCreated new IBDGraph %d\n\n", i); */

   /* Running through connections, start by reading a line of input.          */
   /* key = Hk_FromInt (edge);                                                */
   while (getKeyFromFileStream (&key, edge, fp)
          && fscanf (fp, "%d %d %d", &ibd0, &ibd1, &changes) == 3)
   {
      /* printf ("Building graph %d\n", i); */
      /* printf ("Read %s %d %d %d\n", edge, ibd0, ibd1, changes); */
      cur_range_min = mr_graph_min;     /* Check.                             */

      /* Checking if the edge just read is already in the graph.              */
      /* If so, start a new graph.                                            */
      if (IBDGraphContainsEdgeWithHashKey (ibd_graph, key))
      {
         Igl_Give (ibd_graphs, ibd_graph);
         i++;
         mr_graph_min = 0;
         ibd_graph = NewIBDGraph (i);
         /* printf ("** ** **"); */
      }

      /* Adds an edge, with one set of connections, to the graph.  Edge is    */
      /* the edge_name.  Ibd0 is the base_node.                               */
      e = IBDGraphEdgeByHashKey (ibd_graph, key);
      cur_node = IBDGraphNodeByNumber (ibd_graph, ibd0);

      /* Call function to add the connection.                                 */
      IBDGraph_Connect (ibd_graph, e, cur_node, cur_range_min, 1);
      /* printf ("  connecting initial: 0 1\n"); */
      cur_range_min = 1;
      cur_node = IBDGraphNodeByNumber (ibd_graph, ibd1);

      if (changes > 0)                  /* Checking if changes to manage.     */
      {
         /* If there are changes, run through and update the graph.           */
         for (k = 0; k < changes; k++)
         {
            fscanf (fp, "%d %d", &change_pos, &ibd);
            /* printf ("  Changes %d %d\n", change_pos, ibd); */

            /* Call function to add the connection.                           */
            IBDGraph_Connect (ibd_graph, e, cur_node, cur_range_min,
                              change_pos);
            /* printf ("  connecting changes: %d %d\n", cur_range_min, change_pos); */
            cur_range_min = change_pos;
            cur_node = IBDGraphNodeByNumber (ibd_graph, ibd);
         }
      }

      /* No changes in this line so just add connection.                      */
      IBDGraph_Connect (ibd_graph, e, cur_node, cur_range_min,
                        Mr_Plus_Infinity ());

      /* Reading paired line to above.                                        */

      if (verifyKeyAndAdvanceStream (edge, fp) && fscanf (fp, "%d %d %d", &ibd0,
                                                          &ibd1, &changes) == 3)
      {
         /* printf ("Read %s %d %d %d\n", edge, ibd0, ibd1, changes); */

         /* Adds an edge, with one set of connections, to the graph.  Edge is */
         /* the edge_name.  Ibd0 is the base_node.                            */
         e = IBDGraphEdgeByHashKey (ibd_graph, key);
         cur_node = IBDGraphNodeByNumber (ibd_graph, ibd0);

         cur_range_min = mr_graph_min;  /* Check.                             */

         /* Call function to add the connection.                              */
         IBDGraph_Connect (ibd_graph, e, cur_node, cur_range_min, 1);
         /* printf ("  connecting initial: 0 1\n"); */
         cur_range_min = 1;
         cur_node = IBDGraphNodeByNumber (ibd_graph, ibd1);

         if (changes > 0)               /* Checking if changes to manage.     */
         {
            /* If there are changes, update the graph.                        */
            for (k = 0; k < changes; k++)
            {
               fscanf (fp, "%d %d", &change_pos, &ibd);
               /* printf ("  Changes %d %d\n", change_pos, ibd); */

               /* Call function to add the connection.                        */
               IBDGraph_Connect (ibd_graph, e, cur_node, cur_range_min,
                                 change_pos);
               /* printf ("  connecting changes: %d %d\n", cur_range_min, change_pos); */
               cur_range_min = change_pos;
               cur_node = IBDGraphNodeByNumber (ibd_graph, ibd);
            }
         }

         /* No changes in this line so just add connection.                   */
         IBDGraph_Connect (ibd_graph, e, cur_node, cur_range_min,
                           Mr_Plus_Infinity ());
         /* printf ("  Connecting: %d %d\n", cur_range_min, (int) Mr_Plus_Infinity ()); */
      }
      else
      {
         printf ("  ERROR: Parsing error.");
      }
   }

   Igl_Give (ibd_graphs, ibd_graph);
   fclose (fp);

   return;
}

/******************************************************************************/
/*                                                                            */
/* Print a single marker interval.                                            */
/*                                                                            */
/******************************************************************************/

void PrintInterval (markertype ml, markertype mu)
{
   printf (" [%ld,", ml);

   if (mu != MARKER_PLUS_INFTY)
      printf ("%ld),", mu);
   else
      printf ("inf)");

   return;
}

/******************************************************************************/
/*                                                                            */
/* Allocate an mi_list_node, inserting the newly allocated node in the linked */
/* list of mi_list_node's, keeping these in ascending order of the starting   */
/* marker number in the marker interval defined within the markerinfo object  */
/* passed as the single argument to this subroutine.                          */
/*                                                                            */
/******************************************************************************/

MI_LIST_NODE_PTR alloc_mi_list_node (MI_PTR mi_ptr)
{
   MI_LIST_NODE_PTR new_node, cur_node;

   new_node = malloc (sizeof (MI_LIST));
   alloc_chk (WHERE, new_node);

   new_node->p_mi = mi_ptr;
   new_node->p_next = NULL;

   if (!first_mi_list_node ||
       first_mi_list_node->p_mi->r.start > mi_ptr->r.start)
   {
      new_node->p_next = first_mi_list_node; /* Put at the start of the list  */
      first_mi_list_node = new_node;    /*      (may be only node).           */
   }
   else
   {
      /* Insert the newly allocated node in or add to the end of the list.    */
      cur_node = first_mi_list_node;
      while (cur_node->p_next
             && cur_node->p_next->p_mi->r.start < mi_ptr->r.start)
         cur_node = cur_node->p_next;

      new_node->p_next = cur_node->p_next;
      cur_node->p_next = new_node;
   }

   return new_node;
}

/******************************************************************************/
/*                                                                            */
/* Free the memory used for the linked list of mi_list_node's.                */
/*                                                                            */
/******************************************************************************/

void free_mi_list (void)
{
   MI_LIST_NODE_PTR p_miln, p_nxt;

   p_miln = first_mi_list_node;
   while (p_miln)
   {
      p_nxt = p_miln->p_next;
      free (p_miln);
      p_miln = p_nxt;
   }
   first_mi_list_node = NULL;

   return;
}
