/******************************************************************************/
/*                                                                            */
/* The find_num_indivs subroutine finds the total number of individuals in    */
/* the unprocessed IBD graphs input file produced by gl_auto.  While doing    */
/* this the subroutine also initializes the comp_all_indivs vector to the     */
/* total number of individuals read from the unprocessed IBDgraphs file for   */
/* each component.  The find_num_indivs subroutine also saves the beginning   */
/* line numbers for each component read from the unprocessed IBDgraphs input  */
/* file in the comp_first_line_num vector.                                    */
/*                                                                            */
/******************************************************************************/

void find_num_indivs (int *tot_num_indivs)
{
   FILE *iXtra;                         /* Unprocessed IBDgraphs input file.  */

   char  comp_first_nam[MAXNAMLEN + 1] = "";
   char  prev_nam[MAXNAMLEN + 1] = "";
   char  next_nam[MAXNAMLEN + 1];

   char  nxt_lin[MAXLINLEN + 2];
   char *p_rec;
   char *p_tok;

   long  line_num;                      /* Line number of line just read from */
   /*                                   the unprocessed IBDgraphs input file. */

   long  difference;                    /* The difference between the current */
   /*                                      line number and the start of the   */
   /*                                      current component's line number.   */

   int   first_realization;             /* Binary flag, initialized to TRUE,  */
   /*                                      then set to FALSE after the first  */
   /*                                      realization of IBDgraphs for each  */
   /*                                      component have been read.          */

   int   curr_comp_no;                  /* Current component number.          */

   int   comp_ndx;                      /* Current component subscript.       */
   int   comp_no;                       /* Local copy of component number.    */
   int   comp_vec_size;                 /* Reallocated length of any integer  */
   /*                                      vectors with comp_no elements.     */
   int   ped_ndx;                       /* Pedigree and nghd[] subscript.     */

   int   remainder;                     /* Result of the remainder function.  */

   int   num_indivs;                    /* A local variable for keeping the   */
   /*                                      total number of individuals found  */
   /*                                      in the unprocessed IBDgraphs file. */

   int   err;

   if (!(iXtra = open_fil (XTRA)))      /* Open the IBDgraphs input file.     */
      give_up (ERR_qit);

   comp_first_line_num = (long *) malloc ((n_compnts + 1) * sizeof (long));
   alloc_chk (WHERE, comp_first_line_num);
   comp_first_line_num[0] = 1;

   comp_first_indiv_index = (int *) malloc (n_compnts * sizeof (int));
   alloc_chk (WHERE, comp_first_indiv_index);
   memset (comp_first_indiv_index, INVALID_INDEX, n_compnts * sizeof (int));

   comp_all_indivs = alloc_int1d (n_compnts);
   comp_included = alloc_int1d (n_compnts);

   first_realization = TRUE;
   input_num_graphs = INVALID_INDEX;
   curr_comp_no = UNDEF;
   line_num = 0;

   /* Read until end-of-file from the unprocessed IBDgraphs input_file.       */
   while ((p_rec = fgets (nxt_lin, MAXLINLEN + 1, iXtra)) != NULL)
   {
      ++line_num;                       /* Increment the line number.         */
      if (DebugLevel == 6)
         printf (" Input IBD graphs file line number %ld: %s\n", line_num,
                 p_rec);

      p_tok = strtok (p_rec, WHITESPACE);
      s_read_name (p_tok, next_nam, &err);
      if (err != 1)
      {
         printf (" Error interpreting %s as a name\n", p_tok);
         continue;
      }

      /* If it is not already determined, try to determine what the component */
      /* number is for the current component.                                 */

      if (curr_comp_no == UNDEF)
      {
         ped_ndx = get_posit (next_nam);
         if (ped_ndx != INVALID_INDEX)
         {
            curr_comp_no = nghd[ped_ndx].component;
            comp_included[curr_comp_no - 1] = TRUE;
         }
      }

      if (first_realization)
      {
         if (strcmp (comp_first_nam, "") == 0)
         {
            /* This is the first line of the unprocessed IBDgraphs file.      */
            strcpy (comp_first_nam, next_nam);
            comp_ndx = 0;
            comp_first_line_num[comp_ndx] = line_num;
         }
         else if (end_of_realization (comp_first_nam, prev_nam, next_nam,
                                      curr_comp_no) == TRUE)
         {
            /* We have just finished reading the first realization of IBD     */
            /* graphs for this component from the unprocessed IBDgraphs file. */

            /* Compute the number of lines needed for this first realization  */
            /* for this component input from the unprocessed IBDgraphs file.  */

            lines_per_graph = line_num - comp_first_line_num[comp_ndx];

            first_realization = FALSE;
         }
      }

      difference = line_num - comp_first_line_num[comp_ndx];

      /* Check to make sure that if the current line is an even numbered line */
      /* the name on this line matches the name on the previous line.         */

      if ((line_num % 2) == 0)
      {
         if (strcmp (prev_nam, next_nam) != 0)
         {
            printf (" In the input IBD graphs file the name found on line");
            printf (" %ld: \"%s\",\n              ", (line_num - 1), prev_nam);
            printf (" does not match the name found on line %ld:", line_num);
            printf (" \"%s\".\n", next_nam);
            free_DGL_vectors ();
            exit (94);
         }
      }

      /* Otherwise if this is an odd numbered line and this is not the first  */
      /* realization of IBDgraphs from the unprocessed IBDgraphs file, then   */
      /* check to see if the current name matches the first name read in for  */
      /* the current component.  So long as this is not the the very first    */
      /* IBDgraph read in, if the two names are not the same, check to see if */
      /* the current line should be ending the data for the current           */
      /* component.                                                           */

      else if (difference > 0)
      {
         if (!first_realization && (strcmp (comp_first_nam, next_nam) != 0))
         {
            remainder = difference % lines_per_graph;
            if (remainder == 0)
            {
               /* Since next_nam did not match comp_first_nam, it looks like  */
               /* this line is the beginning of data for a new component.     */

               strcpy (comp_first_nam, next_nam);
               ++comp_ndx;
               comp_first_line_num[comp_ndx] = line_num;

               /* Check for the case where the number of components of data   */
               /* in the unprocessed IBDgraphs input file is greater than the */
               /* number of components in the user's input individuals file.  */

               if (comp_ndx >= n_compnts)
               {
                  comp_no = comp_ndx + 1;
                  comp_first_line_num =
                     realloc (comp_first_line_num, (comp_no + 1) * sizeof
                              (long));

                  comp_vec_size = comp_no * sizeof (int);
                  comp_first_indiv_index = realloc (comp_first_indiv_index,
                                                    comp_vec_size);
                  comp_all_indivs = realloc (comp_all_indivs, comp_vec_size);
                  comp_included = realloc (comp_included, comp_vec_size);
               }

               /* If we have just finished reading the first component's data */
               /* compute the number of IBDgraphs input for this component.   */

               if (input_num_graphs == INVALID_INDEX)
                  input_num_graphs = (line_num - 1) / lines_per_graph;

               /* Calculate the total number of individuals found reading the */
               /* previous component's IBDgraphs from the input file.         */

               comp_all_indivs[comp_ndx - 1] =
                  compute_num_indivs (difference, comp_ndx);

               first_realization = TRUE;
               curr_comp_no = UNDEF;
            }
         }
      }

      strcpy (prev_nam, next_nam);
   }

   fclose (iXtra);

   /* If there is only one realization in the unprocessed IBDgraphs input     */
   /* file the lines_per_graph variable will not have been set, so set it     */
   /* here.                                                                   */

   if (first_realization)
      lines_per_graph = (line_num + 1) - comp_first_line_num[comp_ndx];

   /* If there is only one component in the unprocessed IBDgraphs input file  */
   /* we need to compute the number of IBDgraphs input for the component.     */

   if (input_num_graphs == INVALID_INDEX)
      input_num_graphs = line_num / lines_per_graph;

   /* Store the number of components found in the unprocessed IBDgraphs input */
   /* file in the gl_lods program "global" variable, num_comp_input.  This    */
   /* variable will be needed to know how many elements there are in the      */
   /* comp_first_indiv_index and comp_all_indivs vectors.                     */

   num_comp_input = comp_ndx + 1;

   /* Calculate the total number of individuals found reading the final       */
   /* component's IBDgraphs from the IBDgraphs file.                          */

   difference = (line_num + 1) - comp_first_line_num[comp_ndx];
   comp_all_indivs[comp_ndx] = compute_num_indivs (difference, num_comp_input);

   /* Store one greater than the total number of lines in the unprocessed     */
   /* IBDgraphs input file in the final entry in comp_first_line_num to       */
   /* prevent the code determining component number from going past the end   */
   /* of this vector.                                                         */

   comp_first_line_num[num_comp_input] = (line_num + 1);

   /* Compute the total number of individuals found in the unprocessed        */
   /* IBDgraphs input file to return to the caller.                           */

   /* Also, in this same loop, determine the DGL_index subscript for the      */
   /* first individual in each component.                                     */

   num_indivs = 0;
   for (comp_ndx = 0; comp_ndx < num_comp_input; ++comp_ndx)
   {
      comp_first_indiv_index[comp_ndx] = num_indivs;
      num_indivs += comp_all_indivs[comp_ndx];
   }

   *tot_num_indivs = num_indivs;

   return;
}

/******************************************************************************/
/*                                                                            */
/* The compute_num_indivs subroutine attempts to use the first argument to    */
/* determine how many individuals belong to the component selected by the     */
/* second argument.  If there is no missing data in the unprocessed IBDgraphs */
/* input file the number computed by this subroutine should be correct.       */
/*                                                                            */
/* The first argument is the number of lines read from the unprocessed        */
/* IBDgraphs input file for the component indicated by the second argument.   */
/*                                                                            */
/******************************************************************************/

int compute_num_indivs (long difference, int comp_no)
{
   long  denominator;                   /* A temporary for the denominator.   */

   int   remainder;                     /* Result of the remainder function.  */
   int   result;

   remainder = difference % input_num_graphs;
   if (remainder > 0)
   {
      /* The total number of lines read from the unprocessed IBDgraphs input  */
      /* file for each component should be an even multiple of the number of  */
      /* IBDgraphs found for the first component.                             */

      printf (" ******************** WARNING ******************** WARNING");
      printf (" ********************\n");
      printf (" *                                                        ");
      printf ("                    *\n");
      printf (" * The number of lines in the input IBD graphs file for co");
      printf ("mponent %3d is not  *\n", comp_no);
      printf (" * an even multiple of the number of IBDgraphs found for t");
      printf ("he first realization. *\n");
      printf (" *                                                        ");
      printf ("                    *\n");
      printf (" * The number of individuals shown below is probably not c");
      printf ("orrect so any error *\n");
      printf (" * messages pertaining to this component are likely to be ");
      printf ("misleading.         *\n");
      printf (" *                                                        ");
      printf ("                    *\n");
      printf (" ******************** WARNING ******************** WARNING");
      printf (" ********************\n\n\n");
   }

   /* Calculate the total number of individuals belonging to the component.   */

   denominator = input_num_graphs * 2;
   result = (int) (difference / denominator);

   return result;
}

/******************************************************************************/
/*                                                                            */
/* The find_max_fgl subroutine finds the maximum founder genome label (FGL)   */
/* used anywhere in the unprocessed IBD graphs input file produced by gl_auto.*/
/*                                                                            */
/******************************************************************************/

void find_max_fgl (int *max_fgl)
{
   FILTAB_PTR ptb;                      /* A file table pointer for files.    */
   FILE *iXtra;                         /* Unprocessed IBDgraphs input file.  */

   char  comp_first_nam[MAXNAMLEN + 1] = "";
   char  prev_nam[MAXNAMLEN + 1] = "";
   char  next_nam[MAXNAMLEN + 1];

   char  nxt_lin[MAXLINLEN + 2];
   char *p_rec;
   char *p_tok;

   long  difference;                    /* The difference between the current */
   /*                                      line number and the start of the   */
   /*                                      current component's line number.   */

   long  line_num;                      /* Line number of line just read from */
   /*                                      the unprocessed IBDgraphs file.    */

   int  *nghd_index;                    /* The entries in this vector are set */
   /*                                      TRUE when the indexed individual   */
   /*                                      is located in the unprocessed      */
   /*                                      IBDgraphs input file.              */

   int   comp_first_realization;        /* Binary flag, initialized to TRUE,  */
   /*                                      then set FALSE after the current   */
   /*                                      component's first realization of   */
   /*                                      IBDgraph's has been read.          */

   int   comp_ndx;                      /* Current component subscript.       */

   int   prev_comp_no;                  /* Previous component number.         */
   int   curr_comp_no;                  /* Current component number.          */
   int   ind_comp_no;                   /* Individual's component number.     */

   int   remainder;                     /* Result of the remainder function.  */

   int   indiv_index;                   /* Individual's DGL_index subscript.  */
   int   curr_index;                    /* Individual's nghd[] subscript.     */

   int   ped_ndx;                       /* Pedigree and nghd[] subscript.     */

   int   cur_fgl;                       /* Current founder genome label.      */
   int   m_fgl;                         /* Maximum FGL found so far.          */

   int   num_cp;                        /* Current graph's number of changes. */
   int   cp;                            /* Current change point.              */
   int   c1;                            /* Subscript to current change point. */

   int   err;

   ptb = get_filent (XTRA);
   if (!(iXtra = fopen (ptb->fname, "r")))
      give_up (ERR_qit);
   if (DebugLevel == 6)
      printf (" Opened unprocessed IBD graphs input file \"%s\"\n\n",
              ptb->fname);

   nghd_index = alloc_int1d (pedsize);

   comp_first_realization = TRUE;
   prev_comp_no = 0;
   comp_ndx = 0;
   line_num = 0;
   m_fgl = 0;

   /* Read until end-of-file from the unprocessed IBDgraphs input_file.       */
   while ((p_rec = fgets (nxt_lin, MAXLINLEN + 1, iXtra)) != NULL)
   {
      ++line_num;                       /* Increment the line number.         */
      if (DebugLevel == 6)
         printf (" Input IBD graphs file line number %ld: %s\n", line_num,
                 p_rec);

      p_tok = strtok (p_rec, WHITESPACE);
      s_read_name (p_tok, next_nam, &err);
      if (err != 1)
      {
         printf (" Error interpreting %s as a name\n", p_tok);
         continue;
      }

      if (comp_first_realization)
      {
         if (strcmp (comp_first_nam, "") == 0)
         {
            /* This is the first line of the unprocessed IBDgraphs file.      */
            strcpy (comp_first_nam, next_nam);
         }
         else if (end_of_realization (comp_first_nam, prev_nam, next_nam,
                                      curr_comp_no) == TRUE)
         {
            /* We have just finished reading one realization of IBDgraphs.    */
            comp_first_realization = FALSE;
         }
      }

      /* Determine which component we are currently working on.               */

      find_comp_no (line_num, &curr_comp_no);

      /* If this is an odd numbered line, determine whether or not we are     */
      /* beginning a new component.                                           */

      if ((line_num % 2) != 0)
      {
         if (prev_comp_no < curr_comp_no)
         {
            strcpy (comp_first_nam, next_nam);
            comp_first_realization = TRUE;
            prev_comp_no = curr_comp_no;
            comp_ndx = curr_comp_no - 1;
            lines_per_graph = comp_all_indivs[comp_ndx] * 2;
         }
      }

      /* Compute the DGL_index subscript for the current individual.          */

      difference = line_num - comp_first_line_num[comp_ndx];
      remainder = difference % lines_per_graph;
      indiv_index = comp_first_indiv_index[comp_ndx] + (remainder / 2);

      /* Compute the individual's nghd[] index for the current individual.    */

      curr_index = get_posit (next_nam);

      if (comp_first_realization)
      {
         /* If the individual's name has not already been copied to the       */
         /* DGL_names vector, then copy this new name there now.              */

         if (strcmp (prev_nam, next_nam) != 0)
            strcpy (DGL_names[indiv_index], next_nam);

         if (curr_index != INVALID_INDEX)
         {
            nghd_index[curr_index] = TRUE;
            if (nghd[curr_index].obs)
               DGL_index[indiv_index] = curr_index;
         }
      }

      else
      {
         /* Check to see if the current individual's name correctly matches   */
         /* the corresponding name from the first realization.  If not, we    */
         /* need to produce an error message before terminating gl_lods.      */

         if (strcmp (next_nam, DGL_names[indiv_index]) != 0)
         {
            printf (" On line %ld of the input IBD graphs file", line_num);
            printf (" the name found: \"%s\",\n               ", next_nam);
            printf ("         does not match the name expected:");
            printf (" \"%s\".\n", DGL_names[indiv_index]);
            free_DGL_vectors ();
            exit (95);
         }
      }

      if (curr_index != INVALID_INDEX)
      {
         ind_comp_no = nghd[curr_index].component;
         if (ind_comp_no < curr_comp_no)
         {
            /* Produce an error message about individual being out of place.  */
            printf (" Encountered individual %s from component", next_nam);
            printf (" %d while reading component", ind_comp_no);
            printf (" %d's IBD graphs\n\n", curr_comp_no);
            free_DGL_vectors ();
            exit (96);
         }

         p_tok = strtok (NULL, WHITESPACE); /* Scan past the null trait FGL.  */
         sscanf (p_tok, " %d", &cur_fgl);

         p_tok = strtok (NULL, WHITESPACE); /* Get first marker locus FGL.    */
         sscanf (p_tok, " %d", &cur_fgl);
         if (cur_fgl > m_fgl)
            m_fgl = cur_fgl;

         p_tok = strtok (NULL, WHITESPACE); /* Get number of change points.   */
         sscanf (p_tok, " %d", &num_cp);

         for (c1 = 0; c1 < num_cp; ++c1) /* For each change point:            */
         {
            p_tok = strtok (NULL, WHITESPACE); /* Scan past marker number.    */
            sscanf (p_tok, " %d", &cp);
            p_tok = strtok (NULL, WHITESPACE); /* Get the replacement FGL.    */
            sscanf (p_tok, " %d", &cur_fgl);
            if (cur_fgl > m_fgl)
               m_fgl = cur_fgl;
         }
      }

      strcpy (prev_nam, next_nam);
   }

   for (ped_ndx = 0; ped_ndx < pedsize; ++ped_ndx) /* Now check nghd_index[]. */
      if ((nghd[ped_ndx].obs == TRUE) && (nghd_index[ped_ndx] == 0))
      {
         printf (" Observed individual %s is not in the", nghd[ped_ndx].self);
         printf (" input IBD graphs file:\n");
         printf ("   The trait data on %s will be ignored (W)\n\n",
                 nghd[ped_ndx].self);
         nghd[ped_ndx].obs = FALSE;
      }

   fclose (iXtra);

   free_1d ((void *) &nghd_index);

   /* Return the largest FGL found in the input file to the caller.           */
   *max_fgl = m_fgl;

   return;
}

/******************************************************************************/
/*                                                                            */
/* The find_comp_no subroutine uses the input line_num passed as the first    */
/* argument to determine which component's input data is currently being read */
/* in.  The component number is returned as the second argument.              */
/*                                                                            */
/******************************************************************************/

void find_comp_no (long line_num, int *comp_no)
{
   int   curr_comp_no;                  /* Current component number.          */
   int   c1;                            /* The comp_first_line_num subscript. */

   curr_comp_no = 0;
   for (c1 = 0; c1 < num_comp_input; ++c1)
   {
      if (line_num < comp_first_line_num[c1])
         break;
      else
         ++curr_comp_no;
   }

   *comp_no = curr_comp_no;

   return;
}

/******************************************************************************/
/*                                                                            */
/* Call the count_num_obs subroutine to allocate and initialize a vector of   */
/* integers, comp_num_obs, where each entry in the vector will be set to the  */
/* number of observed individuals in the corresponding component.  Be aware   */
/* that only components found with useful data (ie, not trios) in the input   */
/* IBD graphs file will have entries in this vector.  In other words, the     */
/* first entry will be for the first component found in the input IBD graphs  */
/* file, the second entry will be for the second component in the file, etc.  */
/* As a result, the subscripts into this vector do not necesssarily match the */
/* component numbers found in the nghd array.                                 */
/*                                                                            */
/******************************************************************************/

void count_num_obs (void)
{
   int   indiv_ndx;                     /* Individual's DGL_index subscript.  */
   int   nghd_ndx;                      /* Individual's nghd subscript.       */

   int   comp_no;                       /* Local copy of component number.    */
   int   comp_ndx;                      /* Current component subscript.       */

   comp_num_obs = alloc_int1d (num_comp_input);

   for (indiv_ndx = 0; indiv_ndx < tot_num_indivs; ++indiv_ndx)
   {
      nghd_ndx = DGL_index[indiv_ndx];
      if (nghd_ndx != INVALID_INDEX)
      {
         comp_no = nghd[nghd_ndx].component;
         comp_ndx = find_comp_ndx (comp_no);
         ++comp_num_obs[comp_ndx];
      }
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* The setup_nghd_indices subroutine returns a ragged array, indivs_in_comp,  */
/* containing the nghd subscripts of each observed individual in the          */
/* corresponding component.  Be aware that only components found with useful  */
/* data (ie, not trios) in the input IBD graphs file will have rows in this   */
/* array.  In other words, the first row will be for the first informative    */
/* component found in the input IBD graphs file, the second row will be for   */
/* the second informative component in the file, etc.  As a result, the row   */
/* subscripts into this array do not necesssarily match the component numbers */
/* found in the nghd array.                                                   */
/*                                                                            */
/* Before setup_nghd_indices is called the count_num_obs subroutine must have */
/* been called to allocate and compute the comp_num_obs vector.  The          */
/* comp_num_obs vector is needed by setup_nghd_indices to pass the counts of  */
/* the number of integers in each row of indivs_in_comp.                      */
/*                                                                            */
/******************************************************************************/

void setup_nghd_indices (int ***indivs_in_comp)
{
   int  *indiv_subscripts;              /* Subscripts into indivs_in_comp.    */

   int   indiv_ndx;                     /* Individual's DGL_index subscript.  */
   int   nghd_ndx;                      /* Individual's nghd subscript.       */

   int   nghd_comp_no;                  /* Component number found in nghd.    */
   int   comp_no;                       /* Local copy of component number.    */
   int   comp_ndx;                      /* Current component subscript.       */

   *indivs_in_comp = ragged_alloc_int2d (num_comp_input, comp_num_obs);

   indiv_subscripts = alloc_int1d (num_comp_input);

   for (indiv_ndx = 0; indiv_ndx < tot_num_indivs; ++indiv_ndx)
   {
      nghd_ndx = DGL_index[indiv_ndx];
      if (nghd_ndx != INVALID_INDEX)
      {
         /* Save nghd_ndx in correct indivs_in_comp row.                       */
         comp_no = nghd[nghd_ndx].component;
         comp_ndx = find_comp_ndx (comp_no);
         (*indivs_in_comp)[comp_ndx][indiv_subscripts[comp_ndx]] = nghd_ndx;
         ++indiv_subscripts[comp_ndx];
      }
   }

   free_1d ((void *) &indiv_subscripts);

   if (DebugLevel == 1)
   {
      for (comp_ndx = 0; comp_ndx < num_comp_input; ++comp_ndx)
      {
         /* Get the nghd component number from the first observed individual  */
         /* belonging to the current component.                               */

         nghd_ndx = (*indivs_in_comp)[comp_ndx][0];
         nghd_comp_no = nghd[nghd_ndx].component;

         printf (" The nghd subscripts of all observed individuals in");
         printf (" component %d are:\n", nghd_comp_no);

         for (indiv_ndx = 0; indiv_ndx < comp_num_obs[comp_ndx]; ++indiv_ndx)
            printf (" %d", (*indivs_in_comp)[comp_ndx][indiv_ndx]);

         printf ("\n\n");
      }

      printf ("\n");
      fflush (stdout);
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* Call the find_comp_ndx subroutine to figure out the subscript to use for   */
/* the component indicated in the argument to this subroutine by reference to */
/* the comp_included vector of binary flags.  If an entry is a 1,  the        */
/* component was included.  Otherwise it wasn't.  If the component number     */
/* passed as the argument to find_comp_ndx is larger than the number of       */
/* components input from the individuals file, there is something wrong in    */
/* gl_lods, so terminate with an appropriate error message.                   */
/*                                                                            */
/******************************************************************************/

int find_comp_ndx (int comp_no)
{
   int   comp_ndx;                      /* Computed component subscript.      */
   int   c1;                            /* Subscript into comp_included.      */

   if (comp_no > n_compnts)
   {
      printf (" Terminal error: An internal error in gl_lods has occurred. ");
      printf (" Passed component number, %d, is larger than the", comp_no);
      printf (" total number of components in nghd: %d\n\n", n_compnts);
      free_DGL_vectors ();
      exit (97);
   }
   else                                 /* Otherwise, comp_no <= n_compnts.   */
   {
      comp_ndx = -1;                    /* Initialize comp_ndx so that we can */
      /*                                   produce a correct subscript.       */

      for (c1 = 0; c1 < comp_no; ++c1)
      {
         if (comp_included[c1] == TRUE)
            ++comp_ndx;
      }
   }

   return comp_ndx;
}

/******************************************************************************/
/*                                                                            */
/* The remove_blank subroutine takes as input a string of characters and if   */
/* the last two characters are ' \n' replaces them with a '\n\0'.  In effect, */
/* this removes any extra trailing blank character from the passed string.    */
/*                                                                            */
/******************************************************************************/

void remove_blank (char *string)
{
   int   x1;                            /* Subscript to the last two          */
   /*                                      characters in the passed string.   */

   x1 = strlen (string);
   if (x1 > 1)
   {
      if (strncmp (&string[x1 - 2], " \n", 2) == 0)
         strncpy (&string[x1 - 2], "\n\0", 2);
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* The get_component_graphs subroutine creates a temporary file containing    */
/* only those IBDgraphs belonging to individuals in the selected component.   */
/*                                                                            */
/* The get_component_graphs subroutine copies all the IBD graphs belonging to */
/* observed individuals in the selected component from the input extra file   */
/* to the output scores file.                                                 */
/*                                                                            */
/* Be careful regarding the comp_no argument passed to this subroutine.  The  */
/* program calling get_component_graphs will be working its way through the   */
/* input unprocessed IBD graphs file wanting get_component_graphs to copy to  */
/* the temporary output file only those IBD graphs belonging to the first     */
/* component in the input file, then the second component found in the input  */
/* file, then the third component found in the input file, and so on.  So the */
/* comp_no argument will take on the values 1, 2, 3, etc.  These values are   */
/* not the component numbers as found in nghd[].                              */
/*                                                                            */
/* The get_component_graphs subroutine has to figure out the actual component */
/* number for the component copied to the temporary file.  It is this latter  */
/* component number which is returned to the caller as get_component_graphs'  */
/* return value.                                                              */
/*                                                                            */
/******************************************************************************/

int get_component_graphs (int comp_no)
{
   FILTAB_PTR ptb;                      /* A file table pointer for files.    */

   FILE *iXtra;                         /* Unprocessed IBDgraphs input file.  */
   FILE *oXtra;                         /* Processed IBDgraphs output file.   */

   char  comp_first_nam[MAXNAMLEN + 1] = "";
   char  prev_nam[MAXNAMLEN + 1] = "";
   char  next_nam[MAXNAMLEN + 1] = "";

   char  nxt_lin[MAXLINLEN + 2];
   char *p_rec;
   char *p_tok;
   char *p_out;

   long  line_num;                      /* Line number of line just read from */
   /*                                   the unprocessed IBDgraphs input file. */

   long  graph_num;                     /* Current IBDgraph number.           */

   int   initialize_origin;             /* Binary flag, initialized to TRUE,  */
   /*                                      then set FALSE after encountering  */
   /*                                      the very first individual in the   */
   /*                                      selected component.                */

   int   comp_first_realization;        /* Binary flag, initialized to TRUE,  */
   /*                                      then set FALSE after the selected  */
   /*                                      component's first realization of   */
   /*                                      IBDgraph's has been read.          */

   int   need_comp_no;                  /* Binary flag, initialized to FALSE, */
   /*                                      then set TRUE each time IBDgraphs  */
   /*                                      belonging to a new component are   */
   /*                                      encountered.  When need_comp_no is */
   /*                                      TRUE, get_component_graphs knows   */
   /*                                      that the nghd_comp_no variable     */
   /*                                      needs to be determined.            */

   int   comp_ind_origin;               /* First DGL_index subscript for the  */
   /*                                      selected component.                */

   int   curr_index;                    /* Individual's nghd[] subscript.     */
   int   indiv_comp;                    /* Individual's component number.     */

   int   indiv_cnt;                     /* Count of observed individuals in   */
   /*                                      selected component.                */

   int   prev_comp_no;                  /* Previous component number.         */
   int   curr_comp_no;                  /* Current component number.          */
   int   comp_ndx;                      /* Current component subscript.       */

   int   nghd_comp_no;                  /* Component number found in nghd[]   */
   /*                                      for the IBDgraphs to be copied to  */
   /*                                      the reduced file.                  */

   int   l_name;                        /* Length of input individual's name. */

   int   err;

   /* Create a temporary file made up of only those IBD graphs belonging to   */
   /* the current component only for any observed individuals.                */

   ptb = get_filent (XTRA);
   if (!(iXtra = fopen (ptb->fname, "r")))
      give_up (ERR_qit);
   if (DebugLevel == 6)
      printf (" Opened unprocessed IBD graphs file \"%s\"\n", ptb->fname);

   ptb = get_filent (OXTR);
   if (!(oXtra = fopen (ptb->fname, "w")))
      give_up (ERR_qit);
   if (DebugLevel == 6)
      printf (" Opened reduced IBD graphs file \"%s\"\n\n\n", ptb->fname);

   /* Calculate the number of lines to read in from the unprocessed IBDgraphs */
   /* input file for each realization of IBDgraphs.                           */

   comp_first_realization = FALSE;
   initialize_origin = TRUE;
   need_comp_no = FALSE;

   prev_comp_no = 0;
   nghd_comp_no = 0;
   graph_num = 0;
   indiv_cnt = 0;
   line_num = 0;

   /* Read until end-of-file from the input file, copying only those lines    */
   /* belonging to individuals in the selected component to the output file.  */

   while ((p_rec = fgets (nxt_lin, MAXLINLEN + 1, iXtra)) != NULL)
   {
      ++line_num;                       /* Increment the line number.         */
      if (DebugLevel == 6)
         printf (" Input IBD graphs file line number %ld: %s\n", line_num,
                 p_rec);

      s_read_name (p_rec, next_nam, &err);
      if (err != 1)
      {
         printf (" Error interpreting %s as a name\n", p_rec);
         continue;
      }

      /* Determine if we still need to initialize the origin for the selected */
      /* component.                                                           */

      if (initialize_origin)
      {
         /* Determine which component's unprocessed IBDgraphs input file data */
         /* is being read in based on the line_num we are currently at.       */

         find_comp_no (line_num, &curr_comp_no);

         /* If the origin is still not initialized and the current component  */
         /* number is greater than the largest component number found in the  */
         /* input individuals file, then there were no IBDgraphs in the       */
         /* unprocessed IBDgraphs input file for the selected component.      */

         if (curr_comp_no > n_compnts)
         {
            printf (" There are no IBD graphs belonging to component");
            printf (" %d in the IBD graphs input file\n", comp_no);
            break;
         }

         /* If the current IBDgraph belongs to a new component, re-initialize */
         /* any variables we need to change for the new component.            */

         if (prev_comp_no < curr_comp_no)
         {
            comp_ind_origin = 0;
            for (comp_ndx = 0; comp_ndx < (curr_comp_no - 1); ++comp_ndx)
               comp_ind_origin += comp_all_indivs[comp_ndx];

            prev_comp_no = curr_comp_no;
            need_comp_no = TRUE;
            comp_ndx = curr_comp_no - 1;
            lines_per_graph = comp_all_indivs[comp_ndx] * 2;
         }

         /* If the named individual is the very first individual belonging to */
         /* the selected component in the unprocessed IBDgraphs input file,   */
         /* then initialize those variables reflecting the beginning of the   */
         /* selected component.                                               */

         if (curr_comp_no == comp_no)
         {
            strcpy (comp_first_nam, next_nam);
            comp_first_realization = TRUE;
            initialize_origin = FALSE;
         }
      }

      /* Check to see if all the IBDgraphs for the selected component have    */
      /* been copied to the temporary output file.                            */

      if (line_num >= comp_first_line_num[comp_ndx + 1])
         break;

      if (comp_first_realization)
      {
         /* If we are copying the first realization of IBD graphs for the     */
         /* selected component and if this is the first line of input data    */
         /* for this individual and this is not the first line of data for    */
         /* the selected component, then we have just finished reading the    */
         /* first realization of IBDgraphs for the component.                 */

         if (strcmp (prev_nam, next_nam) != 0)
         {
            if ((line_num > comp_first_line_num[comp_ndx]) &&
                (strcmp (comp_first_nam, next_nam) == 0))
               comp_first_realization = FALSE;
         }
      }

      /* If the name just input from the unprocessed IBD graphs input file    */
      /* matches the first name belonging to the selected component and this  */
      /* is the first of the two lines containing this individual's FGL's,    */
      /* then this is the beginning of the next IBD graph.                    */

      if (end_of_realization (comp_first_nam, prev_nam, next_nam,
                              curr_comp_no) == TRUE)
      {
         /* Check to see if the number of IBDgraphs read so far from the      */
         /* unprocessed IBDgraphs input file is still less than the maximum   */
         /* number of IBDgraphs the user has requested.  If it is, then we    */
         /* need to copy this next IBDgraph to the temporary output file.     */

         if (graph_num < use_num_graphs)
            ++graph_num;
         else
            break;
      }

      /* Determine whether or not the individual named in the most recently   */
      /* read line of input data is to be included in the reduced DGL graphs  */
      /* file.  In doing so, get the subscript into the nghd vector for the   */
      /* individual.                                                          */

      curr_index = check_inclusion (next_nam);

      if (curr_index != INVALID_INDEX)
      {
         if (need_comp_no && line_num >= comp_first_line_num[comp_no - 1])
         {
            nghd_comp_no = nghd[curr_index].component;
            need_comp_no = FALSE;
         }

         indiv_comp = nghd[curr_index].component;
         if (indiv_comp == nghd_comp_no)
         {
            if (strcmp (next_nam, nghd[curr_index].self) == 0)
            {
               /* If we are copying the first realization of IBD graphs for   */
               /* the selected component and if this is the first line of     */
               /* input data for this individual, then increment the count of */
               /* individuals included in the current component's output      */
               /* extra file.                                                 */

               if (comp_first_realization)
               {
                  if (strcmp (prev_nam, next_nam) != 0)
                  {
                     if (nghd[curr_index].obs)
                     {
                        /* Then this is a new observed individual for the     */
                        /* selected component.                                */
                        ++indiv_cnt;
                     }
                  }
               }

               /* Replace the individual's name with the individual's index.  */
               p_tok = p_rec + strspn (p_rec, WHITESPACE);
               l_name = strcspn (p_tok, WHITESPACE);
               p_out = p_tok + l_name;
               remove_blank (p_out);
               fprintf (oXtra, " %d", curr_index);
               fprintf (oXtra, "%s", p_out);
            }
         }
      }

      strcpy (prev_nam, next_nam);
   }

   fclose (iXtra);
   fclose (oXtra);

   if (comp_num_obs[comp_no - 1] != indiv_cnt)
   {
      printf (" Terminal error: An internal error in gl_lods has occurred. ");
      printf (" The number of observed\n individuals supposed to be copied");
      printf (" to the reduced IBD graphs file for component");
      printf (" %d\n does not equal the number of observed", nghd_comp_no);
      printf (" individuals actually copied.\n\n");
      free_DGL_vectors ();
      exit (98);
   }

   return nghd_comp_no;
}

/******************************************************************************/
/*                                                                            */
/* The end_of_realization subroutine is called to determine whether or not    */
/* during the reading of the unprocessed IBDgraphs input file the end of a    */
/* realization of ibd_graphs has been reached.                                */
/*                                                                            */
/******************************************************************************/

int end_of_realization (char *comp_first_nam, char *prev_nam, char *next_nam,
                        int curr_comp_no)
{
   int   next_ndx;                      /* Individual's nghd[] subscript.     */
   int   next_comp_no;                  /* Individual's component number.     */

   /* If prev_nam and next_nam are the same we are working on the second line */
   /* of an ibd_graph, so this can not be the end of a realization.           */

   if (strcmp (prev_nam, next_nam) == 0)
      return FALSE;

   /* If the current component is not being included in the gl_lods run, then */
   /* the subroutine which called end_of_realization would not want an end of */
   /* realization to be returned just because the next individual's component */
   /* number is different, so there is no need to check for this.             */

   if (curr_comp_no != UNDEF && comp_included[curr_comp_no - 1] == TRUE)
   {
      /* Obtain the next individual's nghd[] subscript.                       */

      next_ndx = get_posit (next_nam);

      if (next_ndx != INVALID_INDEX)
      {
         /* Get the component number to which the next individual belongs.    */

         next_comp_no = nghd[next_ndx].component;

         /* If the current component number passed to this subroutine and the */
         /* next component number aren't the same, then in this case a        */
         /* realization would be ending.                                      */

         if (curr_comp_no != next_comp_no)
            return TRUE;
      }
   }

   /* If the next individual's name matches the name of the first individual  */
   /* in the component then a realization must be ending.                     */

   if (strcmp (comp_first_nam, next_nam) == 0)
      return TRUE;

   /* Otherwise, a realization is not ending.                                 */

   return FALSE;
}

/******************************************************************************/
/*                                                                            */
/* Call the check_inclusion subroutine to determine the subscript for a given */
/* name in the nghd vector for the named individual.  If the named individual */
/* is to be included in the reduced IBD graphs file, the subscript into the   */
/* nghd vector for the name individual is returned.  If the named individual  */
/* is not to be included in the reduced IBD graphs file, the check_inclusion  */
/* subroutine returns INVALID_INDEX.                                          */
/*                                                                            */
/******************************************************************************/

int check_inclusion (const char *name_to_find)
{
   int   indiv_index;                   /* Individual's DGL_index subscript.  */
   int   nghd_index;                    /* Individual's nghd subscript.       */

   indiv_index = get_DGL_posit (name_to_find);
   if (indiv_index != INVALID_INDEX)
      nghd_index = DGL_index[indiv_index];
   else
   {
      /* So do not include the named individual.                              */
      nghd_index = INVALID_INDEX;
   }

   return nghd_index;
}

/******************************************************************************/
/*                                                                            */
/* Call the get_DGL_posit subroutine to determine the subscript for a given   */
/* name in the DGL_names vector.  If the name can be found in DGL_names, the  */
/* subscript to the name is returned.  If the given name cannot be found, the */
/* get_DGL_posit subroutine returns INVALID_INDEX.                            */
/*                                                                            */
/******************************************************************************/

int get_DGL_posit (const char *name_to_find)
{
   int   indiv_index;                   /* Individual's DGL_index subscript.  */

   for (indiv_index = 0; indiv_index < tot_num_indivs; ++indiv_index)
   {
      if (strcmp (name_to_find, DGL_names[indiv_index]) == 0)
         return indiv_index;
   }

   /* When the name was not found in DGL_names, return INVALID_INDEX.         */

   return INVALID_INDEX;
}

/******************************************************************************/
/*                                                                            */
/* The gl_lods_alloc subroutine allocates the storage needed for the arrays   */
/* required so that gl_lods will be able to handle the number of IBD graphs   */
/* passed as the one argument to this subroutine.  Note that the entries in   */
/* these arrays are not initialized.                                          */
/*                                                                            */
/******************************************************************************/

void gl_lods_alloc (long num_graphs)
{
   long  graph_ndx;                     /* Current IBDgraph subscript.        */
   long  mark_ndx;                      /* Current marker subscript.          */

   lses_storage = (LSCORE_ELEM **) malloc (num_graphs * sizeof (LSCORE_ELEM *));
   alloc_chk (WHERE, lses_storage);

   lses_storage[0] =
      (LSCORE_ELEM *) malloc (num_graphs * NMarks * sizeof (LSCORE_ELEM));
   alloc_chk (WHERE, lses_storage[0]);

   p_lse_index = (LSE_PTR **) malloc (num_graphs * sizeof (LSE_PTR *));
   alloc_chk (WHERE, p_lse_index);

   p_lse_index[0] = (LSE_PTR *) malloc (num_graphs * NMarks * sizeof (LSE_PTR));
   alloc_chk (WHERE, p_lse_index[0]);

   class_index = (long **) malloc (num_graphs * sizeof (long *));
   alloc_chk (WHERE, class_index);

   class_index[0] = (long *) malloc (num_graphs * NMarks * sizeof (long));
   alloc_chk (WHERE, class_index[0]);

   for (graph_ndx = 1; graph_ndx < num_graphs; ++graph_ndx)
   {
      lses_storage[graph_ndx] = lses_storage[graph_ndx - 1] + NMarks;
      p_lse_index[graph_ndx] = p_lse_index[graph_ndx - 1] + NMarks;
      class_index[graph_ndx] = class_index[graph_ndx - 1] + NMarks;
   }

   equivs_vector = malloc (NMarks * sizeof (MARK_CLASS));
   alloc_chk (WHERE, equivs_vector);

   /* Initialize the p_lse_index to point to the corresponding                */
   /* lodscore_element within lses_storage.                                   */

   for (graph_ndx = 0; graph_ndx < num_graphs; ++graph_ndx)
      for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
         p_lse_index[graph_ndx][mark_ndx] = &lses_storage[graph_ndx][mark_ndx];

   return;
}

/******************************************************************************/
/*                                                                            */
/* This subroutine is called to free up all storage allocated for the current */
/* component for which log likelihoods and lod scores have just been computed */
/* by the gl_lods main program.                                               */
/*                                                                            */
/******************************************************************************/

void gl_lods_free (long num_graphs)
{
   long  graph_ndx;                     /* Current IBDgraph subscript.        */
   long  mark_ndx;                      /* Current marker subscript.          */

   for (graph_ndx = 0; graph_ndx < num_graphs; ++graph_ndx)
   {
      for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
         free_1d ((void *) &lses_storage[graph_ndx][mark_ndx].loglike);
   }
   free_2d ((void *) &lses_storage);

   free_2d ((void *) &p_lse_index);
   free_2d ((void *) &class_index);

   for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
   {
      free_1d ((void *) &equivs_vector[mark_ndx].num_graphs);
      free_1d ((void *) &equivs_vector[mark_ndx].p_lse);
   }

   free_1d ((void *) &equivs_vector);

   return;
}

/******************************************************************************/
/*                                                                            */
/* Subroutine called by gl_lods to determine the equivalence classes for each */
/* requested marker.  The single argument to this subroutine is a linked list */
/* of IBD graphs constructed by the NewIBDGraphList subroutine.               */
/*                                                                            */
/******************************************************************************/

void det_equivalence_classes (IBDGraphList * ibd_graphs)
{
   IBDGraphEquivalences *ige;           /* IBDGraphEquivalences pointer.      */

   LSE_PTR p_lse;                       /* Working lodscore_element pointer.  */

   long  graph_num;                     /* Current IBDgraph number.           */
   long  graph_ndx;                     /* Current IBDgraph subscript.        */
   long  mark_num;                      /* Current marker number.             */
   long  mark_ndx;                      /* Current marker subscript.          */

   long  num_classes;                   /* Number of classes at marker.       */
   long  num_graphs;                    /* Number of graphs.                  */

   long  c1;                            /* Equivalence class subscript.       */
   long  g1;                            /* Current graph subscript.           */

   for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
   {
      mark_num = MarLod[mark_ndx];
      equivs_vector[mark_ndx].mark_num = mark_num;

      ige = IBDGraphEquivalenceClassesAtMarker (ibd_graphs, mark_num);
      IBDGraphEquivalences_InplaceSort (ige);

      if (DebugLevel == 3)
      {
         printf ("\n Equivalence classes at marker %ld:\n\n", mark_num);
         IBDGraphEquivalences_Print (ige);
         printf ("\n");
      }

      equivs_vector[mark_ndx].num_classes = ige->n_classes;
      num_classes = ige->n_classes;

      equivs_vector[mark_ndx].num_graphs = malloc (num_classes * sizeof (long));
      alloc_chk (WHERE, equivs_vector[mark_ndx].num_graphs);

      equivs_vector[mark_ndx].p_lse = malloc (num_classes * sizeof (LSE_PTR));
      alloc_chk (WHERE, equivs_vector[mark_ndx].p_lse);

      /* For each class initialize and point to a single lodscore_element.    */
      /* Note that the actual IBDgraph number is not retained.  Instead it is */
      /* used to set the class_index matrix.                                  */

      for (c1 = 0; c1 < num_classes; ++c1)
      {
         graph_num = ige->classes[c1].graphs[0]->id;
         graph_ndx = graph_num - 1;

         p_lse = p_lse_index[graph_ndx][mark_ndx];
         initialize_lodscore_element (p_lse);

         equivs_vector[mark_ndx].p_lse[c1] = p_lse;
         ++p_lse->num_refs;
         class_index[graph_ndx][mark_ndx] = c1;

         /* For each remaining graph in this equivalence class point to the   */
         /* first graph's lodscore_element.                                   */

         num_graphs = ige->classes[c1].n_graphs;
         equivs_vector[mark_ndx].num_graphs[c1] = num_graphs;

         for (g1 = 1; g1 < num_graphs; ++g1)
         {
            graph_num = ige->classes[c1].graphs[g1]->id;
            graph_ndx = graph_num - 1;

            p_lse_index[graph_ndx][mark_ndx] = p_lse;
            ++p_lse->num_refs;
            class_index[graph_ndx][mark_ndx] = c1;
         }
      }

      O_DecRef (ige);
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* Initialize a lodscore_element.                                             */
/*                                                                            */
/******************************************************************************/

void initialize_lodscore_element (LSE_PTR p_lse)
{
   p_lse->need_loglike = TRUE;
   p_lse->num_refs = 0;
   p_lse->loglike = NULL;

   return;
}

/******************************************************************************/
/*                                                                            */
/*  For a specified equivalence class, this subroutine collapses all of the   */
/*  lodscore_elements for markers within the equivalence class to a single    */
/*  lodscore_element.                                                         */
/*                                                                            */
/******************************************************************************/

void collapse_lodscore_elements (IBDGraphLocationEquivalences * igeq, long c1)
{
   IBDGraphLocation *igloc;             /* Current IBDgraph location item.    */
   LSE_PTR p_lse;                       /* Lodscore_element pointer.          */

   long  class_size;                    /* Number of graphs in the specified  */
   /*                                      equivalence class.                 */

   long  graph_num;                     /* Current IBDgraph number.           */
   long  mark_num;                      /* First marker in current range.     */
   long  mark_end;                      /* Last marker in current range.      */

   long  g1;                            /* Current graph subscript.           */

   /* Set p_lse to NULL when starting on the next equivalence class.          */
   p_lse = NULL;

   class_size = IBDLocEq_ClassSize (igeq, c1);

   if (DebugLevel == 4 && class_size > 1)
   {
      printf ("%5ld:  %4ld ", c1, class_size);
      for (g1 = 0; g1 < class_size; ++g1)
      {
         igloc = IBDLocEq_ViewItem (igeq, c1, g1);
         graph_num = igloc->graph->id;
         mark_num = igloc->start;
         mark_end = igloc->end;
         printf (" (%ld:%ld-%ld)", graph_num, mark_num, mark_end);
      }
      printf ("\n");
   }

   for (g1 = 0; g1 < class_size; ++g1)
   {
      igloc = IBDLocEq_ViewItem (igeq, c1, g1);
      graph_num = igloc->graph->id;
      mark_num = igloc->start;
      mark_end = igloc->end;

      p_lse = do_single_range (p_lse, graph_num, mark_num, mark_end);
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* Collapse the lodscore_elements already in use for the range passed to this */
/* subroutine to the single lodscore_element in use for the first selected    */
/* marker within this range.                                                  */
/*                                                                            */
/******************************************************************************/

LSE_PTR do_single_range (LSE_PTR p_lse, long graph_num, long mark_num,
                         long mark_end)
{
   long  mark_ndx;                      /* Current marker subscript.          */
   long  graph_ndx;                     /* Current IBDgraph subscript.        */
   long  class_ndx;                     /* Current class subscript.           */

   if (mark_end == MARKER_PLUS_INFTY)
      mark_end = MarLod[NMarks - 1] + 1;

   while (mark_num < mark_end)
   {
      mark_ndx = comp_mark_ndx (mark_num);
      if (mark_ndx >= 0)
      {
         graph_ndx = graph_num - 1;
         if (p_lse == NULL)
         {
            /* This is the first lodscore_element for the equivalence class.  */
            p_lse = p_lse_index[graph_ndx][mark_ndx];
         }
         else
         {
            /* Otherwise, the previously found lodscore_element should be     */
            /* used for this selected marker as well as the earlier one.      */
            --(p_lse_index[graph_ndx][mark_ndx]->num_refs);
            p_lse_index[graph_ndx][mark_ndx] = p_lse;
            ++p_lse->num_refs;
         }

         /* In either case save the pointer to this lodscore_element in       */
         /* equivs_vector[mark_ndx].p_lse[class_ndx].                         */

         class_ndx = class_index[graph_ndx][mark_ndx];
         equivs_vector[mark_ndx].p_lse[class_ndx] = p_lse;
         ++p_lse->num_refs;
      }
      ++mark_num;
   }

   return p_lse;
}

/******************************************************************************/
/*                                                                            */
/* The set_perm_trait subroutine sets up the array of permutations of the     */
/* trait data.                                                                */
/*                                                                            */
/* The arguments passed to set_perm_trait are:                                */
/*                                                                            */
/* int   indiv_cnt;                     Count of observed individuals in the  */
/*                                      selected component.                   */
/* int   nperm;                         Number of permutations needed.        */
/* int * idex;                          Vector containing the subscripts in   */
/*                                      nghd for all the observed individuals */
/*                                      in the selected component.            */
/* double **perms;                      An array in which this subroutine is  */
/*                                      to return the resulting permutations. */
/*                                                                            */
/******************************************************************************/

void set_perm_trait (int indiv_cnt, int nperm, int *idex, double **perms)
{
   int  *tperm;                         /* A local vector for producing the   */
   /*                                      each of the permutations in.       */

   int   indiv_index;                   /* Current individual's subscript.    */
   int   perm_index;                    /* Current permutation's subscript.   */

   tperm = alloc_int1d (indiv_cnt);

   /* Copy the actual data to the first column of the matrix.                 */

   for (indiv_index = 0; indiv_index < indiv_cnt; ++indiv_index)
      perms[indiv_index][0] = nghd[idex[indiv_index]].q_dat[0];

   for (perm_index = 1; perm_index <= nperm; ++perm_index) /* We do mean 1.   */
   {
      /* Copy the indiv_cnt nghd subscripts passed in the idex vector to a    */
      /* local copy of the vector of subscripts for all observed individuals  */
      /* in the selected component.                                           */

      for (indiv_index = 0; indiv_index < indiv_cnt; ++indiv_index)
         tperm[indiv_index] = idex[indiv_index];

      /* Produce a permutation of these indiv_cnt nghd subscripts.            */

      permute (tperm, indiv_cnt);

      /* Put the data into the array according to the permutation.            */

      for (indiv_index = 0; indiv_index < indiv_cnt; ++indiv_index)
         perms[indiv_index][perm_index] = nghd[tperm[indiv_index]].q_dat[0];
   }

   /* The local copy of the indiv_cnt nghd subscripts is no longer needed.    */

   free_1d ((void *) &tperm);

   /* The rest of this subroutine should eventually only be executed if the   */
   /* global DebugLevel variable is set to some specific value (probably 1),  */
   /* since it just outputs the resulting permutation data which has been     */
   /* produced by this subroutine.                                            */

   printf ("\n chk perm data");

   for (indiv_index = 0; indiv_index < indiv_cnt; ++indiv_index)
   {
      printf ("\n  %d", idex[indiv_index]);
      for (perm_index = 0; perm_index <= nperm; ++perm_index)
         printf (" %7.3f", perms[indiv_index][perm_index]);
   }
   printf ("\n\n");
   fflush (stdout);

   return;
}

/******************************************************************************/
/*                                                                            */
/* The prnt_contributions subroutine is called by gl_lods to output the log   */
/* likelihood contributions for all IBD graphs at each requested marker for   */
/* the selected component.                                                    */
/*                                                                            */
/* The arguments passed to prnt_contributions are:                            */
/*                                                                            */
/* FILE *input_file;                    Processed IBD graphs input file.      */
/* FILE *oScorf;                        Output file for the log likelihoods.  */
/* long  num_graphs;                    Number of graphs to use.              */
/* int   num_indivs;                    Reduced number of individuals.        */
/* int   comp_no;                       Selected component number.            */
/* int  *idex;                          Vector containing the subscripts in   */
/*                                      nghd for all the observed individuals */
/*                                      in the selected component.            */
/* double **perm_qdat;                  Array containing permutation data.    */
/*                                                                            */
/******************************************************************************/

void prnt_contributions (FILE * input_file, FILE * oScorf, long num_graphs,
                         int num_indivs, int comp_no, int *idex,
                         double **perm_qdat)
{
   LSE_PTR p_lse;                       /* Working lodscore_element pointer.  */

   long  graph_num;                     /* Current IBDgraph number.           */
   long  graph_ndx;                     /* Current IBDgraph subscript.        */

   long  mark_num;                      /* Current marker number.             */
   long  mark_ndx;                      /* Current marker subscript.          */

   int   perm_ndx;                      /* Current permutation subscript.     */

   int   nc;                            /* Character length of highest        */
   /*                                      numbered component.                */
   int   ng;                            /* Character length of highest        */
   /*                                      numbered ibd graph.                */
   int   nm;                            /* Character length of highest        */
   /*                                      numbered selected marker.          */
   int   np;                            /* Character length of most negative  */
   /*                                      permuted log likelihood value.     */

   printf (" Log likelihoods for %ld ibd graphs at markers\n", num_graphs);
   printf (" ");
   for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
      printf ("  %d", MarLod[mark_ndx]);
   printf ("\n");
   printf (" will be printed to the output scores file\n");
   fflush (stdout);

   /* For each requested marker, call the do_single_lse subroutine to compute */
   /* the log likelihood for the associated lodscore_element.                 */

   for (graph_num = 1; graph_num <= num_graphs; ++graph_num)
   {
      fgl_set (input_file, num_indivs); /* Set the founder genome labels.     */

      if (DebugMain)
      {
         fprintf (stderr, "\nFor IBD graph %ld,\n", graph_num);
         dbgXXGG ();
      }

      for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
      {
         mark_num = equivs_vector[mark_ndx].mark_num;

         if (doing_permuts)
            do_vector_lse (comp_no, graph_num, mark_num, num_indivs, idex,
                           perm_qdat);
         else
            do_single_lse (comp_no, graph_num, mark_num);
      }

      graph_ndx = graph_num - 1;

      /* Compute character lengths needed for the output to the scores file   */
      /* which follows.                                                       */

      nc = intWid (0, n_compnts);
      ng = intWid (0, num_graphs);
      nm = intWid (0, MarLod[NMarks - 1]);

      /* The character width wanted for np is the width of the most negative  */
      /* number generated for any of the permutation log likelihoods.  Here I */
      /* have assumed that no log likelihood will be less than -999.0, so np  */
      /* will be 11.  Code needs to be added to make this code more general.  */

      np = fldWid (-999.0, 0.0, 6);

      /* Print the contributions to log likelihood for each graph at each of  */
      /* the requested markers.                                               */

      for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
      {
         fprintf (oScorf, " %*d", nc, comp_no);
         fprintf (oScorf, " %*ld", ng, graph_num);
         fprintf (oScorf, " %*d", nm, MarLod[mark_ndx]);
         p_lse = p_lse_index[graph_ndx][mark_ndx];

         if (doing_permuts)
            for (perm_ndx = 0; perm_ndx <= NumbPermuts; ++perm_ndx)
               fprintf (oScorf, " %*f", np, p_lse->loglike[perm_ndx]);
         else
            fprintf (oScorf, " %*f", np, p_lse->loglike[0]);

         fprintf (oScorf, "\n");
      }

   }

   printf ("\n");
   fflush (stdout);

   return;
}

/******************************************************************************/
/*                                                                            */
/* Compute the log likelihood for a single lodscore_element.                  */
/*                                                                            */
/******************************************************************************/

void do_single_lse (int comp_no, long graph_num, long mark_num)
{
   LSE_PTR p_lse;                       /* Lodscore_element pointer.          */

   long  mark_ndx;                      /* Current marker subscript.          */
   long  graph_ndx;                     /* Current IBDgraph subscript.        */

   mark_ndx = comp_mark_ndx (mark_num);
   if (mark_ndx >= 0)
   {
      graph_ndx = graph_num - 1;
      p_lse = p_lse_index[graph_ndx][mark_ndx];

      if (p_lse->need_loglike == TRUE)
      {
         /* for now allocate a single double here  */
         p_lse->loglike = malloc (1 * sizeof (double));
         p_lse->loglike[0] = comp_loglike (comp_no, mark_ndx);
         p_lse->need_loglike = FALSE;
      }
   }

   return;
}

/* Version of do_single_lse, to do log-likelihoods for permutations.          */

void do_vector_lse (int comp_no, long graph_num, long mark_num,
                    int indiv_cnt, int *idx, double **permdat)
{
   LSE_PTR p_lse;                       /* Lodscore_element pointer.          */

   long  mark_ndx;                      /* Current marker subscript.          */
   long  graph_ndx;                     /* Current IBDgraph subscript.        */

   int   perm_ndx;                      /* Current permutation's subscript.   */
   int   indiv_ndx;                     /* Current individual's subscript.    */

   mark_ndx = comp_mark_ndx (mark_num);
   if (mark_ndx >= 0)
   {
      graph_ndx = graph_num - 1;
      p_lse = p_lse_index[graph_ndx][mark_ndx];

      if (p_lse->need_loglike == TRUE)
      {
         /* Allocate a vector of doubles for the permutation log likelihoods. */
         p_lse->loglike = malloc ((NumbPermuts + 1) * sizeof (double));

         /* Save the computed log likelihoods as the first row.               */
         p_lse->loglike[0] = comp_loglike (comp_no, mark_ndx);

         for (perm_ndx = 1; perm_ndx <= NumbPermuts; ++perm_ndx)
         {
            for (indiv_ndx = 0; indiv_ndx < indiv_cnt; ++indiv_ndx)
               nghd[idx[indiv_ndx]].q_dat[0] = permdat[indiv_ndx][perm_ndx];
            p_lse->loglike[perm_ndx] = comp_loglike (comp_no, mark_ndx);
            p_lse->need_loglike = FALSE;

            /* Restore the true data to nghd.                                 */
            for (indiv_ndx = 0; indiv_ndx < indiv_cnt; ++indiv_ndx)
               nghd[idx[indiv_ndx]].q_dat[0] = permdat[indiv_ndx][0];
         }
      }
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* Compute the current marker subscript for the marker number passed as the   */
/* argument to this subroutine.  This subroutine returns a -1 if the passed   */
/* marker number is not a selected marker.                                    */
/*                                                                            */
/******************************************************************************/

long comp_mark_ndx (long mark_num)
{
   long  mark_ndx;                      /* Current subscript into MarLod.     */

   for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
   {
      if (mark_num == MarLod[mark_ndx])
         return mark_ndx;
   }

   /* Return a -1 if the marker is not found in MarLod.                       */

   return -1;
}

/******************************************************************************/
/*                                                                            */
/* This is a short subroutine which serves as the gl_lods interface to the    */
/* gen_pen subroutine.  This subroutine returns the log likelihood at the     */
/* selected marker indicated by the second argument to this subroutine.       */
/*                                                                            */
/******************************************************************************/

double comp_loglike (int comp_no, long mark_ndx)
{
   double loglike;

   int   m_origin;                      /* Subscript to the first marker.     */
   int   fgl_ndx;                       /* Subscript to current FGL's.        */

   int  *fgls[2];                       /* Vector of pointers to FGL's.       */
   int   comp_ndx;                      /* Current component subscript.       */
   int   err;

   m_origin = max (NTLocs, 1);
   fgl_ndx = m_origin + mark_ndx;

   fgls[0] = gg0[fgl_ndx];
   fgls[1] = gg1[fgl_ndx];

   comp_ndx = comp_no - 1;
   loglike = gen_pen (comp_ndx, fgls, loci + 0, nFGL, &err);

   return loglike;
}

/******************************************************************************/
/*                                                                            */
/* If the user has provided a component base log likelihood for the component */
/* number indicated by the one argument, the gl_lods_loglikes subroutine will */
/* compute and print component lod scores for the selected markers.           */
/*                                                                            */
/* Alternatively, if the user has requested permutations, a permutation-based */
/* score will be computed.                                                    */
/*                                                                            */
/******************************************************************************/

void gl_lods_loglikes (int comp_no)
{
   LSE_PTR p_lse;                       /* Working lodscore_element pointer.  */

   double *comp_lod;                    /* Vector of log likelihoods.         */
   double base_loglike;                 /* Base-point log P(Y_T).             */

   double delta_prob;                   /* Current likelihood contribution.   */
   double perm_prob;                    /* Current permutation contribution.  */
   double probability;                  /* Current marker's likelihood.       */
   double normalizer;                   /* Current marker's permutation norm. */
   double base_prob;                    /* Computational base for perm lods.  */

   long  mark_ndx;                      /* Current marker subscript.          */

   long  num_classes;                   /* Number of classes at marker.       */
   long  num_graphs;                    /* Number of graphs.                  */

   long  c1;                            /* Equivalence class subscript.       */

   int   comp_ndx;                      /* Current component subscript.       */
   int   iperm;                         /* Index for permutations.            */

   if (Base_loglikes != NULL)
   {
      /* Allocate a vector for keeping each selected marker's lod score.      */
      comp_lod = alloc_dbl1d (NMarks);

      /* Find the base log likelihood for the indicated component number.     */
      comp_ndx = comp_no - 1;
      base_loglike = Base_loglikes[comp_ndx] * log (10.0);

      for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
      {
         probability = 0.0;             /* Reset total probability to 0.0.    */

         num_classes = equivs_vector[mark_ndx].num_classes;
         for (c1 = 0; c1 < num_classes; ++c1)
         {
            num_graphs = equivs_vector[mark_ndx].num_graphs[c1];
            p_lse = equivs_vector[mark_ndx].p_lse[c1];

            delta_prob = exp (p_lse->loglike[0] - base_loglike);
            probability += (num_graphs * delta_prob / use_num_graphs);
         }
         comp_lod[mark_ndx] = log10 (probability);
      }

      prnt_gl_lods (comp_no, comp_lod);
      free_1d ((void *) &comp_lod);
   }
   else if (doing_permuts)
   {
      printf ("\n");
      printf ("\n Permutation lod scores will be computed\n");

      /* Allocate a vector for keeping each selected marker's lod score.      */
      comp_lod = alloc_dbl1d (NMarks);

      /* This is the easy version: each graph's log-likelihood is             */
      /* standardized by the permutation for that graph.                      */

      for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
      {
         probability = 0.0;             /* Reset total probability to 0.0.    */

         num_classes = equivs_vector[mark_ndx].num_classes;
         for (c1 = 0; c1 < num_classes; ++c1)
         {
            num_graphs = equivs_vector[mark_ndx].num_graphs[c1];
            p_lse = equivs_vector[mark_ndx].p_lse[c1];

            delta_prob = 0.0;           /* Reset for sum over permutations.   */
            for (iperm = 1; iperm <= NumbPermuts; ++iperm)
               delta_prob += exp (p_lse->loglike[iperm] - p_lse->loglike[0]);
            delta_prob = NumbPermuts / delta_prob;
            probability += (num_graphs * delta_prob / use_num_graphs);
         }
         comp_lod[mark_ndx] = log10 (probability);
      }

      printf ("\n   Version 1 of permutation lods\n");
      prnt_gl_lods (comp_no, comp_lod);

      fflush (stdout);

      /* Alternatively total actual can be compared to total perms, but then    */
      /* we may have to worry about    over/under-flow                          */

      for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
      {
         probability = 0.0;             /* Reset total probability to 0.0.    */
         normalizer = 0.0;              /* Reset total normalizing termto 0.0. */

         num_classes = equivs_vector[mark_ndx].num_classes;
         p_lse = equivs_vector[mark_ndx].p_lse[0];
         base_prob = p_lse->loglike[0];

         for (c1 = 0; c1 < num_classes; ++c1)
         {
            num_graphs = equivs_vector[mark_ndx].num_graphs[c1];
            p_lse = equivs_vector[mark_ndx].p_lse[c1];

            delta_prob = exp (p_lse->loglike[0] - base_prob);
            probability += (num_graphs * delta_prob / use_num_graphs);

            for (iperm = 1; iperm <= NumbPermuts; ++iperm)
            {
               perm_prob = exp (p_lse->loglike[iperm] - base_prob);
               normalizer += (num_graphs * perm_prob /
                              (NumbPermuts * use_num_graphs));
            }
         }
         comp_lod[mark_ndx] = log10 (probability) - log10 (normalizer);
      }

      printf ("\n   Version 2 of permutation lods\n");
      fflush (stdout);
      prnt_gl_lods (comp_no, comp_lod);
      free_1d ((void *) &comp_lod);
   }
   else
   {
      fprintf (stderr, "\n");
      fprintf (stderr, " gl_lods is not able to provide lod scores unless");
      fprintf (stderr, " the user provides a base log\n likelihood using");
      fprintf (stderr, " the \"set base log likelihood X1 ...\" parameter");
      fprintf (stderr, " statement (W)\n\n");
   }

   return;
}

/******************************************************************************/
/*                                                                            */
/* The following subroutine contains code merged from the prntLodSHead, the   */
/* prntLodScores and the prntLodSPos subroutines found in the print_lods.c    */
/* file, which is part of the Markers library.  The prnt_gl_lods subroutine   */
/* is specifically tailored for the gl_lods program.  This subroutine outputs */
/* marker names for the requested markers only, followed by the computed lod  */
/* score contributions at each marker.  No traits are involved.               */
/*                                                                            */
/******************************************************************************/

void prnt_gl_lods (int comp_no, double *lod_scores)
{
   long  mark_ndx;                      /* Current marker subscript.          */

   int   nCol1;                         /* Width of marker names column.      */

   nCol1 = LocNamLen > 13 ? LocNamLen : 13;

   printf ("\n Component %d lod scores found by gl_lods:\n\n", comp_no);
   printf (" %*s   %11s%10s\n\n", nCol1, "Marker number", "LodScore", "StdErr");
   fflush (stdout);

   for (mark_ndx = 0; mark_ndx < NMarks; ++mark_ndx)
   {
      printf (" %*s  ", nCol1, MarkNam[mark_ndx]);
      printf (" %11.4f\n", lod_scores[mark_ndx]);
   }

   printf ("\n");

   return;
}

/******************************************************************************/
/*                                                                            */
/* The free_DGL_vectors subroutine frees up the vectors allocated just once   */
/* by the gl_lods main program, but not any of the storage (re)allocated for  */
/* by gl_lods for each component.  It should only be called at the end of     */
/* gl_lods or if gl_lods is ending with a terminal error.                     */
/*                                                                            */
/******************************************************************************/

void free_DGL_vectors (void)
{
   free_1d ((void *) &comp_first_line_num);
   free_1d ((void *) &comp_first_indiv_index);
   free_1d ((void *) &comp_all_indivs);
   free_1d ((void *) &comp_included);
   free_1d ((void *) &comp_num_obs);
   free_2d ((void *) &DGL_names);
   free_1d ((void *) &DGL_index);

   return;
}
