/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */
/* $Id: alloc_array.c,v 1.3.2.7 2015/06/16 20:20:08 slewis Exp $ */

#include "stuff.h"

int  *alloc_int1d (int dim1)
{
   int  *array1d;

   /* Allocate memory for a one dimensional array (vector) of integers.       */
   /* This single dimension will consist of the actual integer vector.        */
   array1d = (int *) malloc (dim1 * sizeof (int));
   alloc_chk (WHERE, array1d);

   /* Initialize all of the actual integer data to zeroes.                    */
   memset (array1d, 0, dim1 * sizeof (int));

   return array1d;
}

int **alloc_int2d (int dim1, int dim2)
{
   int **array2d;

   int   x1;

   /* Allocate memory for a two dimensional array of integers.                */
   /* The first dimension will consist of pointers to the second dimension.   */
   array2d = (int **) malloc (dim1 * sizeof (int *));
   alloc_chk (WHERE, array2d);

   /* The second dimension will consist of the actual integer vectors.        */
   array2d[0] = (int *) malloc (dim1 * dim2 * sizeof (int));
   alloc_chk (WHERE, array2d[0]);

   /* Fill in the pointers in the first dimension so that they point to the   */
   /* actual integer vectors which make up the second dimension.              */
   for (x1 = 1; x1 < dim1; ++x1)
      array2d[x1] = array2d[x1 - 1] + dim2;

   /* Initialize all of the actual integer data to zeroes.                    */
   memset (array2d[0], 0, dim1 * dim2 * sizeof (int));

   return array2d;
}

int ***alloc_int3d (int dim1, int dim2, int dim3)
{
   int ***array3d;
   int  *next_addr;

   int   x1;
   int   x2;

   /* Allocate memory for a three dimensional array of integers.              */
   /* The first dimension will consist of pointers to the second dimension.   */
   array3d = (int ***) malloc (dim1 * sizeof (int **));
   alloc_chk (WHERE, array3d);

   /* The second dimension will consist of pointers to the third dimension.   */
   array3d[0] = (int **) malloc (dim1 * dim2 * sizeof (int *));
   alloc_chk (WHERE, array3d[0]);

   /* Fill in the pointers in the first dimension so that they point to the   */
   /* pointers which make up the second dimension.                            */
   for (x1 = 1; x1 < dim1; ++x1)
      array3d[x1] = array3d[x1 - 1] + dim2;

   /* The third dimension will consist of the actual integer vectors.         */
   array3d[0][0] = (int *) malloc (dim1 * dim2 * dim3 * sizeof (int));
   alloc_chk (WHERE, array3d[0][0]);

   /* Fill in the pointers in the second dimension so that they point to the  */
   /* actual integer vectors which make up the third dimension.               */
   next_addr = array3d[0][0];
   for (x1 = 0; x1 < dim1; ++x1)
   {
      for (x2 = 0; x2 < dim2; ++x2)
      {
         array3d[x1][x2] = next_addr;
         next_addr += dim3;
      }
   }

   /* Initialize all of the actual integer data to zeroes.                    */
   memset (array3d[0][0], 0, dim1 * dim2 * dim3 * sizeof (int));

   return array3d;
}

/******************************************************************************/
/*                                                                            */
/* The ragged_alloc_int2d subroutine is a slightly modified version of the    */
/* alloc_int2d subroutine located in the alloc_array.c file.  This version    */
/* uses the last argument, ragged_dim2, which needs to be a one-dimensional   */
/* integer vector containing the respective counts of integers which need to  */
/* be allocated (and initialized) as the final ragged dimension of the        */
/* resulting array.                                                           */
/*                                                                            */
/******************************************************************************/

int **ragged_alloc_int2d (int dim1, int *ragged_dim2)
{
   int **array2d;
   int  *next_addr;

   int   int_tot_cnt;                   /* Total count of integers allocated. */
   int   x1;

   /* Allocate memory for a two dimensional array of integers.                */
   /* The first dimension will consist of pointers to the second dimension.   */
   array2d = (int **) malloc (dim1 * sizeof (int *));
   alloc_chk (WHERE, array2d);

   /* Calculate the total count of integers which need to be allocated.       */
   int_tot_cnt = 0;
   for (x1 = 0; x1 < dim1; ++x1)
      int_tot_cnt += ragged_dim2[x1];

   /* The second dimension will consist of the actual integer vectors.        */
   array2d[0] = (int *) malloc (int_tot_cnt * sizeof (int));
   alloc_chk (WHERE, array2d[0]);

   /* Fill in the pointers in the first dimension so that they point to the   */
   /* actual integer vectors which make up the second dimension.              */
   next_addr = array2d[0];
   for (x1 = 0; x1 < dim1; ++x1)
   {
      array2d[x1] = next_addr;
      next_addr += ragged_dim2[x1];
   }

   /* Initialize all of the actual integer data to zeroes.                    */
   memset (array2d[0], 0, int_tot_cnt * sizeof (int));

   return array2d;
}

/******************************************************************************/
/*                                                                            */
/* The ragged_alloc_int3d subroutine is a slightly modified version of the    */
/* alloc_int3d subroutine located in the alloc_array.c file.  This version    */
/* uses the last argument, ragged_dim3, which needs to be a two-dimensional   */
/* integer array containing the respective counts of integers which need to   */
/* be allocated (and initialized) as the final ragged dimension of the        */
/* resulting array.                                                           */
/*                                                                            */
/******************************************************************************/

int ***ragged_alloc_int3d (int dim1, int dim2, int **ragged_dim3)
{
   int ***array3d;
   int  *next_addr;

   int   int_tot_cnt;                   /* Total count of integers allocated. */
   int   x1;
   int   x2;

   /* Allocate memory for a three dimensional array of integers.              */
   /* The first dimension will consist of pointers to the second dimension.   */
   array3d = (int ***) malloc (dim1 * sizeof (int **));
   alloc_chk (WHERE, array3d);

   /* The second dimension will consist of pointers to the third dimension.   */
   array3d[0] = (int **) malloc (dim1 * dim2 * sizeof (int *));
   alloc_chk (WHERE, array3d[0]);

   /* Fill in the pointers in the first dimension so that they point to the   */
   /* pointers which make up the second dimension.                            */
   for (x1 = 1; x1 < dim1; ++x1)
      array3d[x1] = array3d[x1 - 1] + dim2;

   /* Calculate the total count of integers which need to be allocated.       */
   int_tot_cnt = 0;
   for (x1 = 0; x1 < dim1; ++x1)
      for (x2 = 0; x2 < dim2; ++x2)
         int_tot_cnt += ragged_dim3[x1][x2];

   /* The third dimension will consist of the actual integer vectors.         */
   array3d[0][0] = (int *) malloc (int_tot_cnt * sizeof (int));
   alloc_chk (WHERE, array3d[0][0]);

   /* Fill in the pointers in the second dimension so that they point to the  */
   /* actual integer vectors which make up the third dimension.               */
   next_addr = array3d[0][0];
   for (x1 = 0; x1 < dim1; ++x1)
      for (x2 = 0; x2 < dim2; ++x2)
      {
         array3d[x1][x2] = next_addr;
         next_addr += ragged_dim3[x1][x2];
      }

   /* Initialize all of the actual integer data to zeroes.                    */
   memset (array3d[0][0], 0.0, int_tot_cnt * sizeof (int));

   return array3d;
}

double *alloc_dbl1d (int dim1)
{
   double *array1d;

   /* Allocate memory for a one dimensional array (vector) of double          */
   /* precision floating point numbers.  This single dimension will consist   */
   /* of the actual double precision floating point number vector.            */
   array1d = (double *) malloc (dim1 * sizeof (double));
   alloc_chk (WHERE, array1d);

   /* Initialize all of the actual double precision floating point numbers to */
   /* zeroes.                                                                 */
   memset (array1d, 0.0, dim1 * sizeof (double));

   return array1d;
}

double **alloc_dbl2d (int dim1, int dim2)
{
   double **array2d;

   int   x1;

   /* Allocate memory for a two dimensional array of double precision         */
   /* floating point numbers.  The first dimension will consist of pointers   */
   /* to the second dimension.                                                */
   array2d = (double **) malloc (dim1 * sizeof (double *));
   alloc_chk (WHERE, array2d);

   /* The second dimension will consist of the actual double precision        */
   /* floating point number vectors.                                          */
   array2d[0] = (double *) malloc (dim1 * dim2 * sizeof (double));
   alloc_chk (WHERE, array2d[0]);

   /* Fill in the pointers in the first dimension so that they point to the   */
   /* actual double precision floating point number vectors which make up the */
   /* second dimension.                                                       */
   for (x1 = 1; x1 < dim1; ++x1)
      array2d[x1] = array2d[x1 - 1] + dim2;

   /* Initialize all of the actual double precision floating point numbers to */
   /* zeroes.                                                                 */
   memset (array2d[0], 0.0, dim1 * dim2 * sizeof (double));

   return array2d;
}

double ***alloc_dbl3d (int dim1, int dim2, int dim3)
{
   double ***array3d;
   double *next_addr;

   int   x1;
   int   x2;

   /* Allocate memory for a three dimensional array of double precision       */
   /* floating point numbers.  The first dimension will consist of pointers   */
   /* to the second dimension.                                                */
   array3d = (double ***) malloc (dim1 * sizeof (double **));
   alloc_chk (WHERE, array3d);

   /* The second dimension will consist of pointers to the third dimension.   */
   array3d[0] = (double **) malloc (dim1 * dim2 * sizeof (double *));
   alloc_chk (WHERE, array3d[0]);

   /* Fill in the pointers in the first dimension so that they point to the   */
   /* pointers which make up the second dimension.                            */
   for (x1 = 1; x1 < dim1; ++x1)
      array3d[x1] = array3d[x1 - 1] + dim2;

   /* The third dimension will consist of the actual double precision         */
   /* floating point number vectors.                                          */
   array3d[0][0] = (double *) malloc (dim1 * dim2 * dim3 * sizeof (double));
   alloc_chk (WHERE, array3d[0][0]);

   /* Fill in the pointers in the second dimension so that they point to the  */
   /* actual double precision floating point number vectors which make up the */
   /* third dimension.                                                        */
   next_addr = array3d[0][0];
   for (x1 = 0; x1 < dim1; ++x1)
      for (x2 = 0; x2 < dim2; ++x2)
      {
         array3d[x1][x2] = next_addr;
         next_addr += dim3;
      }

   /* Initialize all of the actual double precision floating point numbers to */
   /* zeroes.                                                                 */
   memset (array3d[0][0], 0.0, dim1 * dim2 * dim3 * sizeof (double));

   return array3d;
}

/******************************************************************************/
/*                                                                            */
/* The ragged_alloc_dbl2d subroutine is a slightly modified version of the    */
/* alloc_dbl2d subroutine located in the alloc_array.c file.  This version    */
/* uses the last argument, ragged_dim2, which needs to be a one-dimensional   */
/* integer vector containing the respective counts of double precision        */
/* floating point numbers which need to be allocated (and initialized) as     */
/* the final ragged dimension of the resulting array.                         */
/*                                                                            */
/******************************************************************************/

double **ragged_alloc_dbl2d (int dim1, int *ragged_dim2)
{
   double **array2d;
   double *next_addr;

   int   dbl_tot_cnt;                   /* Total count of double precision    */
   /*                                      floating point numbers allocated.  */
   int   x1;

   /* Allocate memory for a two dimensional array of double precision         */
   /* floating point numbers.  The first dimension will consist of pointers   */
   /* to the second dimension.                                                */
   array2d = (double **) malloc (dim1 * sizeof (double *));
   alloc_chk (WHERE, array2d);

   /* Calculate the total count of double precision floating point numbers    */
   /* which need to be allocated.                                             */
   dbl_tot_cnt = 0;
   for (x1 = 0; x1 < dim1; ++x1)
      dbl_tot_cnt += ragged_dim2[x1];

   /* The second dimension will consist of the actual double precision        */
   /* floating point number vectors.                                          */
   array2d[0] = (double *) malloc (dbl_tot_cnt * sizeof (double));
   alloc_chk (WHERE, array2d[0]);

   /* Fill in the pointers in the first dimension so that they point to the   */
   /* actual double precision floating point number vectors which make up the */
   /* second dimension.                                                       */
   next_addr = array2d[0];
   for (x1 = 0; x1 < dim1; ++x1)
   {
      array2d[x1] = next_addr;
      next_addr += ragged_dim2[x1];
   }

   /* Initialize all of the actual double precision floating point numbers to */
   /* zeroes.                                                                 */
   memset (array2d[0], 0.0, dbl_tot_cnt * sizeof (double));

   return array2d;
}

/******************************************************************************/
/*                                                                            */
/* The ragged_alloc_dbl3d subroutine is a slightly modified version of the    */
/* alloc_dbl3d subroutine located in the alloc_array.c file.  This version    */
/* uses the last argument, ragged_dim3, which needs to be a two-dimensional   */
/* integer array containing the respective counts of double precision         */
/* floating point numbers which need to be allocated (and initialized) as     */
/* the final ragged dimension of the resulting array.                         */
/* This subroutine was moved to alloc_array.c in June of 2015, so that it     */
/* would be available anywhere in Morgan.                                     */
/*                                                                            */
/******************************************************************************/

double ***ragged_alloc_dbl3d (int dim1, int dim2, int **ragged_dim3)
{
   double ***array3d;
   double *next_addr;

   int   dbl_tot_cnt;                   /* Total count of double precision    */
   /*                                      floating point numbers allocated.  */
   int   x1;
   int   x2;

   /* Allocate memory for a three dimensional array of double precision       */
   /* floating point numbers.  The first dimension will consist of pointers   */
   /* to the second dimension.                                                */
   array3d = (double ***) malloc (dim1 * sizeof (double **));
   alloc_chk (WHERE, array3d);

   /* The second dimension will consist of pointers to the third dimension.   */
   array3d[0] = (double **) malloc (dim1 * dim2 * sizeof (double *));
   alloc_chk (WHERE, array3d[0]);

   /* Fill in the pointers in the first dimension so that they point to the   */
   /* pointers which make up the second dimension.                            */
   for (x1 = 1; x1 < dim1; ++x1)
      array3d[x1] = array3d[x1 - 1] + dim2;

   /* Calculate the total count of double precision floating point numbers    */
   /* which need to be allocated.                                             */
   dbl_tot_cnt = 0;
   for (x1 = 0; x1 < dim1; ++x1)
      for (x2 = 0; x2 < dim2; ++x2)
         dbl_tot_cnt += ragged_dim3[x1][x2];

   /* The third dimension will consist of the actual double precision         */
   /* floating point number vectors.                                          */
   array3d[0][0] = (double *) malloc (dbl_tot_cnt * sizeof (double));
   alloc_chk (WHERE, array3d[0][0]);

   /* Fill in the pointers in the second dimension so that they point to the  */
   /* actual double precision floating point number vectors which make up the */
   /* third dimension.                                                        */
   next_addr = array3d[0][0];
   for (x1 = 0; x1 < dim1; ++x1)
      for (x2 = 0; x2 < dim2; ++x2)
      {
         array3d[x1][x2] = next_addr;
         next_addr += ragged_dim3[x1][x2];
      }

   /* Initialize all of the actual double precision floating point numbers to */
   /* zeroes.                                                                 */
   memset (array3d[0][0], 0.0, dbl_tot_cnt * sizeof (double));

   return array3d;
}

char *alloc_char1d (int dim1)
{
   char *array1d;

   /* Allocate memory for a one dimensional array (vector) of characters.     */
   /* This single dimension will consist of the actual character string.      */
   array1d = (char *) malloc (dim1 * sizeof (char));
   alloc_chk (WHERE, array1d);

   /* Initialize all of the actual character string data to zeroes.           */
   memset (array1d, 0, dim1 * sizeof (char));

   return array1d;
}

char **alloc_char2d (int dim1, int dim2)
{
   char **array2d;

   int   x1;

   /* Allocate memory for a two dimensional array of character strings.       */
   /* The first dimension will consist of pointers to the second dimension.   */
   array2d = (char **) malloc (dim1 * sizeof (char *));
   alloc_chk (WHERE, array2d);

   /* The second dimension will consist of the actual character string        */
   /* vectors.                                                                */
   array2d[0] = (char *) malloc (dim1 * dim2 * sizeof (char));
   alloc_chk (WHERE, array2d[0]);

   /* Fill in the pointers in the first dimension so that they point to the   */
   /* actual character string vectors which make up the second dimension.     */
   for (x1 = 1; x1 < dim1; ++x1)
      array2d[x1] = array2d[x1 - 1] + dim2;

   /* Initialize all of the actual character string data to zeroes.           */
   memset (array2d[0], 0, dim1 * dim2 * sizeof (char));

   return array2d;
}

char ***alloc_char3d (int dim1, int dim2, int dim3)
{
   char ***array3d;
   char *next_addr;

   int   x1;
   int   x2;

   /* Allocate memory for a three dimensional array of character strings.     */
   /* The first dimension will consist of pointers to the second dimension.   */
   array3d = (char ***) malloc (dim1 * sizeof (char **));
   alloc_chk (WHERE, array3d);

   /* The second dimension will consist of pointers to the third dimension.   */
   array3d[0] = (char **) malloc (dim1 * dim2 * sizeof (char *));
   alloc_chk (WHERE, array3d[0]);

   /* Fill in the pointers in the first dimension so that they point to the   */
   /* pointers which make up the second dimension.                            */
   for (x1 = 1; x1 < dim1; ++x1)
      array3d[x1] = array3d[x1 - 1] + dim2;

   /* The third dimension will consist of the actual character string         */
   /* vectors.                                                                */
   array3d[0][0] = (char *) malloc (dim1 * dim2 * dim3 * sizeof (char));
   alloc_chk (WHERE, array3d[0][0]);

   /* Fill in the pointers in the second dimension so that they point to the  */
   /* actual character string vectors which make up the third dimension.      */
   next_addr = array3d[0][0];
   for (x1 = 0; x1 < dim1; ++x1)
      for (x2 = 0; x2 < dim2; ++x2)
      {
         array3d[x1][x2] = next_addr;
         next_addr += dim3;
      }

   /* Initialize all of the actual character string data to zeroes.           */
   memset (array3d[0][0], 0, dim1 * dim2 * dim3 * sizeof (char));

   return array3d;
}

void free_1d (void **array)
{
   if (*array)
      free (*array);

   *array = NULL;
   return;
}

void free_2d (void ***array)
{
   if (*array)
   {
      if (**array)
         free (**array);
      free (*array);
   }

   *array = NULL;
   return;
}

void free_3d (void ****array)
{
   if (*array)
   {
      if (**array)
      {
         if (***array)
            free (***array);
         free (**array);
      }
      free (*array);
   }

   *array = NULL;
   return;
}

void free_4d (void *****array)
{
   if (*array)
   {
      if (**array)
      {
         if (***array)
         {
            if (****array)
               free (****array);
            free (***array);
         }
         free (**array);
      }
      free (*array);
   }

   *array = NULL;
   return;
}

/******************************************************************************/
/* Routines to allocate int and dbl vectors transferred here: EAT May 2006.   */
/* These are used by Liping's routines that were in MulMsample.               */
/******************************************************************************/

void al_int_vector (int **array, int dim1)
{
   *array = malloc (dim1 * sizeof (int));
   alloc_chk (WHERE, *array);

   return;
}

void al_double_vector (double **array, int dim1)
{
   *array = malloc (dim1 * sizeof (double));
   alloc_chk (WHERE, *array);

   return;
}
