/* MORGAN: Copyright 1991-2005: Univ. of Washington. All rights reserved. */
/* You may copy, use, modify, but NOT redistribute, modified or otherwise.*/
/* See: http://www.stat.washington.edu/thompson/Genepi/license.shtml      */

/*
 NAME
    write_lod_score, write_batch_means, alloc_mc_eval, free_mc_eval,
    update_batch

 SYNOPSIS

    #include "twoqtl.h"
    void write_lod_score (mc_eval_t *mc_eval);
    void write_batch_means (mc_eval_t *mc_eval, int batch_no);

    mc_eval_t *alloc_mc_eval (int n_pos);
    void free_mc_eval (mc_eval_t *mc_eval);

    void update_batch (double value, int start, double *foo_p, double *max_p);

 DESCRIPTION

    write_lod_score computes log probability and mc standard error, using
    the method of batch means and writes them to the standard output file
    (called once after MCMC is done).

    write_batch_means writes batch means into file "batch_means"
    (called when each batch is complete).

    alloc_mc_eval allocates MC_eval and free_mc_eval frees it.

    update_batch accumulates batch sum, called once per iteration,
    and stores into two pieces (foo and max) for stable computation.

 SEE ALSO
    twoqtl.h for the definition of struct mc_eval_t.

 AUTHOR
    Written by Yun Ju Sung.

 HISTORY
    Documented.                                 Apr 21, 2005
    Write batches while running MCMC.           Aug 17, 2005
    Change the code in var_lod.                 Aug 19, 2005

*/

#include "twoqtl.h"

static FILTAB_PTR oXtra_pft;
static FILE *oXtra = NULL;

/* nCol1 is a static variable needed for formatting column headings and data  */
/* columns.  It is initialized in the print_top subroutine.                   */
static int nCol1;

void write_lod_score (mc_eval_t * mc_eval)
{
   double **lod_comp, **var_comp, **lod_all, **var_all;
   double mean_null, var_null, mean_alt, var_alt;
   int   i1, i2, i3;

/* Allocate storage for the two dimensional double precision arrays used by   */
/* this subroutine.                                                           */

   lod_comp = alloc_dbl2d (n_loci_1, n_loci_2);
   var_comp = alloc_dbl2d (n_loci_1, n_loci_2);
   lod_all = alloc_dbl2d (n_loci_1, n_loci_2);
   var_all = alloc_dbl2d (n_loci_1, n_loci_2);

/* Print two blank lines before beginning the output of lod_scores estimated  */
/* for the first component.                                                   */
   printf ("\n\n");

   for (i1 = 0; i1 < n_compnts; ++i1)
   {
      mean_null = mean_batch (mc_eval + i1, 0, 0, 0);

      if (NBtch > 1)
      {
         var_null = var_batch (mc_eval + i1, 0, 0, 0, mean_null);
         var_null /= NBtch * exp (mean_null * 2);
      }

      for (i2 = 0; i2 < n_loci_1; ++i2)
      {
         for (i3 = 0; i3 < n_loci_2; ++i3)
         {
            mean_alt = mean_batch (mc_eval + i1, i2, i3, 1);

            lod_comp[i2][i3] = mean_alt;
            lod_all[i2][i3] += mean_alt - mean_null;

            if (NBtch > 1)
            {
               var_alt = var_batch (mc_eval + i1, i2, i3, 1, mean_alt);
               var_alt /= NBtch * exp (2 * mean_alt);

               var_comp[i2][i3] = var_alt;
               var_all[i2][i3] += var_alt + var_null;
            }
         }
      }

      print_top (0, i1);
      print_null (mean_null, var_null);
      print_lodsd (lod_comp, var_comp);
      printf ("\n");
   }

/* Now print lod_score estimates and MC sd for the entire pedigree.           */

   print_top (1, 0);
   print_lodsd (lod_all, var_all);

/* Free the no longer needed storage previously allocated in this subroutine. */

   free_2d ((void *) &lod_comp);
   free_2d ((void *) &var_comp);
   free_2d ((void *) &lod_all);
   free_2d ((void *) &var_all);

   return;
}

/* Log of mean of batch means.                                                */

double mean_batch (mc_eval_t * mc_eval, int loci_1, int loci_2, int alt)
{
   double foo_n, max_n, value;
   mc_eval_t eval;
   int   i1;

   for (i1 = 0; i1 < NBtch; ++i1)
   {
      eval = mc_eval[n_compnts * i1];

      if (alt == 1)
         value = eval.max_alt[loci_1][loci_2]
            + log (eval.foo_alt[loci_1][loci_2]);
      else
         value = eval.max_null + log (eval.foo_null);

      update_batch (value, i1 == 0, &foo_n, &max_n);
   }

   value = max_n + log (foo_n / NBtch);

   return value;
}

/* Sample variance of batch means.                                            */

double var_batch (mc_eval_t * mc_eval, int loci_1, int loci_2, int alt,
                  double mean)
{
   double value;                        /* Current batch mean.                */
   double result;
   mc_eval_t eval;
   int   i1;

   result = 0;
   for (i1 = 0; i1 < NBtch; ++i1)
   {
      eval = mc_eval[n_compnts * i1];

      if (alt == 1)
         value = eval.foo_alt[loci_1][loci_2]
            * exp (eval.max_alt[loci_1][loci_2]);
      else
         value = eval.foo_null * exp (eval.max_null);

      value -= exp (mean);
      result += value * value;
   }
   result /= NBtch - 1;

   return result;
}

void print_top (int lod_flag, int comp_no)
{
   if (lod_flag == 1)
      printf ("# Lod score estimates and MC sd for entire pedigree:\n\n");
   else
      printf ("# Log (base 10) likelihood estimate and MC sd "
              "for component %d:\n\n", comp_no + 1);

   nCol1 = LocNamLen > 11 ? LocNamLen : 11;

   printf ("# Index %9s", "TLoc1");
   if (n_loci_2 != 1)
      printf ("   %9s", "TLoc2");

   if (lod_flag == 1)
      printf ("  %11s%10s\n\n", "LodScore", "StdErr");
   else
      printf ("  %11s%10s\n\n", "LogLike", "StdErr");

   return;
}

void print_null (double mean_null, double var_null)
{
   printf ("    0 %11s", "unlinked");
   if (n_loci_2 != 1)
      printf (" %11s", "unlinked");
   printf ("  %11.4f%10.4f\n", mean_null / log (10.0),
           sqrt (var_null) / log (10.0));

   return;
}

void print_lodsd (double **lod, double **var)
{
   int   index, i1, i2;

   index = 0;
   for (i1 = 0; i1 < n_loci_1; ++i1)
   {
      for (i2 = 0; i2 < n_loci_2; ++i2) /* Convert into log base 10.          */
      {
         printf (" %4d %11.3f ", ++index, LocPos[i1].m1dist[MAL]);
         if (n_loci_2 != 1)
            printf ("%11.3f ", LocPos[i2].m1dist[MAL]);
         printf (" %11.4f%10.4f\n", lod[i1][i2] / log (10.0),
                 sqrt (var[i1][i2]) / log (10.0));
      }
   }
   printf ("\n");

   return;
}

void write_batch_means (mc_eval_t * mc_eval, int batch_no)
{
   double *foo_alt, *max_alt;
   double value;
   mc_eval_t *eval;
   int   i1, i2, i3;

   if (batch_no == 0)
   {
      oXtra_pft = get_filent (OXTR);
      if (oXtra_pft != NULL && oXtra_pft->fname)
         oXtra = fopen (oXtra_pft->fname, "w");

      if (oXtra)
      {
         fprintf (oXtra,
                  "# %d batches, %d QTL, %d components, %d alternative loci\n\n"
                  "# In each batch, likelihood estimates for components are "
                  "separated by blank\n# lines. For each component, the "
                  "likelihood under the null comes first,\n"
                  "# followed by an array of the likelihood under the "
                  "alternative (1-D array\n"
                  "# for 1 QTL and 2-D array for 2 QTL).\n\n",
                  NBtch, NQtl, n_compnts, n_loci_1);
      }
   }

   if (oXtra)
      fprintf (oXtra, "# likelihood from batch %d\n", batch_no + 1);

   for (i1 = 0; i1 < n_compnts; ++i1)
   {
      eval = mc_eval + i1;
      eval->foo_null /= (BatchSize * MultiplierForNull);
      if (oXtra)
      {
         value = eval->foo_null * exp (eval->max_null);
         fprintf (oXtra, "%14.9e\n", value);
      }

      for (i2 = 0; i2 < n_loci_1; ++i2)
      {
         foo_alt = eval->foo_alt[i2];
         max_alt = eval->max_alt[i2];

         for (i3 = 0; i3 < n_loci_2; ++i3)
         {
            foo_alt[i3] /= BatchSize;
            if (oXtra)
            {
               value = foo_alt[i3] * exp (max_alt[i3]);
               fprintf (oXtra, "%14.9e", value);
               if (i3 + 1 < n_loci_2)
                  fprintf (oXtra, " ");
            }
         }
         if (oXtra)
            fprintf (oXtra, "\n");
      }
      if (oXtra)
         fprintf (oXtra, "\n");
   }

   if (oXtra)
   {
      fflush (oXtra);

      if (batch_no == NBtch - 1)
         fclose (oXtra);
   }

   return;
}

void update_batch (double value, int start, double *foo_p, double *max_p)
{
   double max_before;
   double max_now;

   if (start)
   {
      *foo_p = 1.0;
      *max_p = value;
   }
   else
   {
      max_before = *max_p;
      max_now = max (max_before, value);

      *foo_p *= exp (max_before - max_now);
      *foo_p += exp (value - max_now);
      *max_p = max_now;
   }

   return;
}

mc_eval_t *alloc_mc_eval (int n_loci)
{
   mc_eval_t *eval;
   int   i1;

   /* Initialize three global variables used in the importance sampling done  */
   /* for the Monte Carlo approximation in the simulate () subroutine.        */

   mc_eval_length = NBtch * n_compnts;
   n_loci_1 = n_loci;
   n_loci_2 = (NQtl == 1) ? 1 : n_loci;

   eval = (mc_eval_t *) malloc (mc_eval_length * sizeof (mc_eval_t));
   alloc_chk (WHERE, eval);

   for (i1 = 0; i1 < mc_eval_length; ++i1)
      eval[i1] = setup_mc_eval ();

   return eval;
}

mc_eval_t setup_mc_eval (void)
{
   mc_eval_t eval;

   eval.foo_alt = alloc_dbl2d (n_loci_1, n_loci_2);
   eval.max_alt = alloc_dbl2d (n_loci_1, n_loci_2);

   eval.foo_null = 0.0;
   eval.max_null = 0.0;

   eval.n_loci_1 = n_loci_1;
   eval.n_loci_2 = n_loci_2;

   return eval;
}

void free_mc_eval (mc_eval_t * mc_eval)
{
   int   i1;

   for (i1 = 0; i1 < mc_eval_length; ++i1)
   {
      free_2d ((void *) &mc_eval[i1].foo_alt);
      free_2d ((void *) &mc_eval[i1].max_alt);
   }
   free (mc_eval);

   return;
}

/*******************************************************************************

ESTIMATE AND VARIANCE
   for estimating log P_lambda - log P_null

   our estimate is

      log hat(P_lambda) - log hat(P_null)

   and its variance is

      (sigma_lambda)^2 / n / (P_lambda)^2 + (sigma_null)^2 / n / (P_null)^2

   where hat(P_lambda) is MCMC sample mean of
     P_lambda (Trait | segregation indicators at current iteration)
   and (sigma_lambda)^2 / n is estimated by the method of batch means
   (explained below).

-----------------------------------------------------------------
[var is corrected from Exp(Y_l / mu_l - Y_0 / mu_0)^2 / n   Aug 18, 2005]

(sketch of proof)

  (1) hat(P_lambda) and hat(P_null) are independent.

  (2) By the CLT,

       sqrt(n) [hat(P_lambda) - P_lambda] --> N(0, sigma^2)

    By the delta method,

       sqrt(n) [log hat(P_lambda) - log P_lambda]
                             --> N(0, sigma^2 / P_lambda^2)
    Thus
       var [log hat(P_lambda)] = sigma^2 / n / P_lambda^2

-----------------------------------------------------------------
THEORY FOR ESTIMATING SD VIA THE METHOD OF BATCH MEAN

 sum(Y_i - bar(Y))^2 / [(NBtch - 1) * NBtch ] estimates sigma^2 / n

where Y_i is the i-th batch mean, NBtch is the number of batches,
and n is the number of iterations of MCMC used for evaluation.
-----------------------------------------------------------------
(sketch of proof)

  for X_1, ...., X_n
      sqrt(n) (bar(X)_n - mu) --> N(0, sigma^2)

  for each batch with batch size b goes to infinity
      sqrt(b) (bar(X)_b - mu) --> N(0, sigma^2)

      var(bar(X)_b) = sigma^2 / b

  letting Y_i = bar(X)_b_i (i-th batch mean),
     sample var of batch means [ sum(Y_i - bar(Y))^2 / (NBtch - 1) ]
     estimates sigma^2 / b.

*******************************************************************************/
